import math
from ....atc40.raw.ch8 import csm as csm_atc40


def Teq(T0, μ, α):
    if μ <= 1.0:
        return T0
    else:
        return T0 * (μ / (1 + α*μ - α))**0.5


def β0(μ, α):
    if μ<1.0:
        return 0.
    else:
        return (2/math.pi)*((μ-1)*(1-α))/(μ*(1+α*μ-α))


# import pandas as pd
# import numpy as np
# import math
from dataclasses import dataclass, field
from typing import List, Dict
from tabulate import tabulate
from ....phd.building_models.model2d.slab_loads import *
from ....codes.greek.eak2000.raw.ch2.seismic_action import spectra
from ....ppp.sections.concrete.typical_sections.rectangular_rc import SectionMaterials
from ....ppp.sections.concrete.typical_sections.rectangular_rc import RectangularConcreteSection
from ....codes.greek.kanepe.assessment.section_properties.rectangular import RectangularKanepe as RectKanepe
from ....ppp.sections.concrete.reinforcement.rectangular_reinforcement import RectangularSectionReinforcement
from ....ppp.sections.concrete.reinforcement.combos import LongReinforcementLayer, TransReinforcementLayer
from ....common.io.output import OutputExtended, OutputString
import pickle
import gzip

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import orm
from sqlalchemy import Table, Column, Integer, String, Date, MetaData, Float, ForeignKey
from sqlalchemy import func

from .sql_classes import Node, NodesCase, \
                Section, SectionsCase, \
                Element, ElementsCase, \
                Slab, SlabsCase, \
                NodeSlabConn, NodeSlabConnCase, \
                BeamSlabConn, BeamSlabConnCase, \
                Material, MaterialsCase, \
                BeamReinforcement, BeamReinforcementsCase, \
                ColumnReinforcement, ColumnReinforcementsCase, \
                ModFactor, ModFactorsCase, \
                BuildingIdentity


@dataclass
class Coordinates:
    id: int
    X: float
    Y: float
    Z: float

    def __init__(self, id, x, y, z):
        self.id = id
        self.X = x
        self.Y = y
        self.Z = z

@dataclass
class Model2d_sql:
    input_data: Dict[str, pd.DataFrame] = field(default_factory=dict)
    misc: Dict[str, object] = field(default_factory=dict)

    # stories_total: int = 0
    # storey_levels: List[float] = field(init=False, repr=False, default_factory=list)
    # storey_heights: List[float] = field(init=False, repr=False, default_factory=list)
    # x_levels: List[float] = field(init=False, repr=False, default_factory=list)
    # master_nodes: List[int] = field(init=False, repr=False, default_factory=list)
    # base_level_nodes: List[int] = field(init=False, repr=False, default_factory=list)

    storey_masses: List[float] = field(init=False, repr=False, default_factory=list)
    fist_mode: List[float] = field(init=False, repr=False, default_factory=list)
    periods: List[float] = field(init=False, repr=False, default_factory=list)
    seismic_static_forces: Dict[str, list] = field(init=False, repr=False, default_factory=dict)
    elastic_gravity_element_forces: Dict[int, list] = field(init=False, repr=False, default_factory=dict)

    reinforcement_beams: Dict[str, dict] = field(init=False, repr=False, default_factory=dict)
    reinforcement_beams_kanepe: Dict[str, RectKanepe] = field(init=False, repr=False, default_factory=dict)
    reinforcement_columns_kanepe: Dict[str, RectKanepe] = field(init=False, repr=False, default_factory=dict)

    logs: OutputExtended = field(init=False, repr=False, default_factory=OutputExtended)

    def __post_init__(self):
        self.excel_index_column = dict()
        self.excel_index_column['Misc'] = 'Building'
        self.excel_index_column['Nodes'] = 'NodeID'
        self.excel_index_column['Slabs'] = 'SlabID'
        self.excel_index_column['NodeSlabConn'] = 'NodeID'
        self.excel_index_column['BeamSlabConn'] = 'ElementID'
        self.excel_index_column['Sections'] = 'SectionName'
        self.excel_index_column['Elements'] = 'ElementID'
        self.excel_index_column['Concrete'] = 'ConcID'
        self.excel_index_column['BeamReinforcement'] = 'ReinforcementID'
        self.excel_index_column['ColumnReinforcement'] = 'ReinforcementID'
        self.excel_index_column['ModFactors'] = 'ElementType'


    def get_eigen(self):
        from .os_input import OpenSeesModel2d
        os = OpenSeesModel2d(self, inelastic_analysis=False)
        os.tcl_make = True
        os.tcl_filename = ''
        os.os_run_elastic()
        self.periods = os.results.elastic_periods
        self.fist_mode = os.results.elastic_eigenvectors[0]
        self.elastic_gravity_element_forces = os.results.elastic_gravity_element_forces

        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Periods - Eigenvectors  ------------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        _log.append('  Mode       Period (sec)')
        for i in range(0, self.stories_total):
            _log.append(f'{i+1:>5} {self.periods[i]:14.4f}')

        _log.append('')
        _log.append('Mode 1 - Horizontal displacements')
        _log.append('  Storey       si')
        for i in range(0, self.stories_total):
            _log.append(f'{i+1:>5} {self.fist_mode[i]:14.4f}')

        _log.append('')
        _log.append('')
        self.logs.outputStrings['eigen'] = OutputString(data=_log)

    def get_seismic_static_forces(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Seismic Static Forces  -------------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        T0 = self.periods[0]
        V0_push = 1000.
        storeys = range(1, self.stories_total+1)
        m = np.array(self.storey_masses)
        s = np.array(self.fist_mode)
        ms = m * s

        forces_push_modal = V0_push*ms / sum(ms)
        forces_push_uni = V0_push*m / sum(m)
        forces_eak = np.zeros(len(m))

        if self.misc['EAK_α'] > 0:
            _eak_ground = self.misc['EAK_ground']
            _eak_α = self.misc['EAK_α']
            _eak_Φd = spectra.Φd(T=T0,
                           α=_eak_α*9.81,
                           γI=1.0,
                           T1=spectra.T1(_eak_ground),
                           T2=spectra.T2(_eak_ground),
                           q=3.5,
                           η=1.0,
                           θ=1.0,
                           β0=2.5)

            V0_eak = sum(m) * _eak_Φd

            if T0>1.0:
                VH_eak = min(0.07*T0*V0_eak, 0.25*V0_eak)
            else:
                VH_eak = 0.0

            forces_eak = (V0_eak-VH_eak) * ms / sum(ms)

            _log.append(f'T0 = {T0:.4f}sec')
            _log.append(f'Φd = {_eak_Φd:.3f}m/sec2')
            _log.append(f'V0_eak = {V0_eak:.2f}kN')
            _log.append(f'VH_eak = {VH_eak:.2f}kN')

            _log.append('')

            self.seismic_static_forces['EPUSHmod'] = forces_push_modal.tolist()
            self.seismic_static_forces['EPUSHuni'] = forces_push_uni.tolist()
            self.seismic_static_forces['EXSTAT'] = forces_eak.tolist()



        _df = pd.DataFrame(data = {'storey': storeys, 'm': m, 's': s, 'm*s':ms,
                                   'E_push_modal': forces_push_modal, 'E_push_uni': forces_push_uni,
                                   'E_eak': forces_eak})




        _log.append(tabulate(_df,
                             headers='keys',
                             tablefmt='pipe',
                             floatfmt=".4f"))

        _log.append('')
        _log.append('')
        self.logs.outputStrings['seismic_static_forces'] = OutputString(data=_log)


    def load_excel_single(self, input_excel_filename):
        # Read the excel file
        xls = pd.ExcelFile(input_excel_filename)
        # Get worksheet names
        for sheet_name in xls.sheet_names:
            self.input_data[sheet_name] = xls.parse(sheet_name).set_index(self.excel_index_column[sheet_name])

    def load_excel_collection(self, input_excel_filename, building_props):
        # Read the excel file
        xls = pd.ExcelFile(input_excel_filename)
        # Get worksheet names
        for sheet_name in xls.sheet_names:
            self.input_data[sheet_name] = \
                xls.parse(sheet_name)[xls.parse(sheet_name).Building ==
                                      building_props[sheet_name]].set_index(self.excel_index_column[sheet_name])



    def prepare(self, get_eigen_from_opensees = False):
        self.process_misc()
        self.process_joint_coordinates()
        self.process_sections()
        self.log_table_elements()
        self.log_table_slabs()
        self.log_node_slab_connectivity()
        self.process_node_slab_connectivity()
        self.log_beam_slab_connectivity()
        self.process_beam_slab_connectivity()
        self.process_storey_masses()

        if get_eigen_from_opensees == True:
            self.get_eigen()
            self.get_seismic_static_forces()


    def prepare_Mtheta(self):
        self.process_beam_reinforcement()
        self.process_column_reinforcement()

    def process_misc(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Process Misc  ----------------------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        self.misc = self.input_data['Misc'].iloc[0].to_dict()

        for key, value in self.misc.items():
            _log.append(f'{key} = {value}')
        _log.append('')
        _log.append('')
        self.logs.outputStrings['misc'] = OutputString(data=_log)

    def process_joint_coordinates(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Process Joint Coordinates  ---------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        self.stories_total = max(self.input_data['Nodes']['Storey'])
        self.storey_levels = self.input_data['Nodes']['Z'].unique().tolist()
        self.x_levels = sorted(self.input_data['Nodes']['X'].unique().tolist(), key=float)
        st_h = [self.storey_levels[i + 1] - self.storey_levels[i] for i in range(0, self.stories_total)]
        self.storey_heights = st_h
        self.master_nodes = [int(self.input_data['Nodes'].query(f'Z=={sl}').first_valid_index())
                             for sl in self.storey_levels][1:]
        self.base_level_nodes = self.input_data['Nodes'].query(f'Z==0').index.tolist()

        _log.append(f'stories total = {self.stories_total}')
        _log.append(f'storey levels: {self.storey_levels}')
        _log.append(f'x levels: {self.x_levels}')
        _log.append(f'storey heights: {self.storey_heights}')
        _log.append(f'master nodes: {self.master_nodes}')
        _log.append(f'base level nodes: {self.base_level_nodes}')
        _log.append('')
        _log.append('--------  Coordinates  --------------------------')
        _log.append('   Node         X      Y (Z)      Storey    Point_Etabs')
        for key, value in self.input_data['Nodes'].to_dict('index').items():
            _log.append(f'{key:>7} {value["X"]:9.2f} {value["Z"]:9.2f} {value["Storey"]:>9} {value["Point_Etabs"]:>12}')
        _log.append('')
        _log.append('')
        self.logs.outputStrings['joints'] = OutputString(data=_log)

    def process_sections(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Process Sections  ------------------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        # *********** Προσθήκη στήλης με συνεργαζόμενα πλάτη *************************
        beffs = []
        for index, row in self.input_data['Sections'].iterrows():
            if row['ElementType'] == 'BEAM':
                bw = row['b']
                hf = self.misc['hf']
                if row['Frame_Loc'] == 'IN':
                    factor = 8.
                else:
                    factor = 3.
                beffs.append(bw + factor * hf)
            else:
                beffs.append(None)
        self.input_data['Sections']['beff'] = beffs

        _log.append(tabulate(self.input_data['Sections'],
                             headers='keys',
                             tablefmt='pipe',
                             floatfmt=".2f"))
        _log.append('')
        _log.append('')
        self.logs.outputStrings['sections'] = OutputString(data=_log)

    def log_table_elements(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Elements Table  --------------------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '',
                tabulate(self.input_data['Elements'],
                         headers='keys',
                         tablefmt='pipe',
                         floatfmt=".2f"), '', '']
        self.logs.outputStrings['elements'] = OutputString(data=_log)

    def log_table_slabs(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Slabs Table  -----------------------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '',
                tabulate(self.input_data['Slabs'],
                         headers='keys',
                         tablefmt='pipe',
                         floatfmt=".2f"), '', '']
        self.logs.outputStrings['slabs'] = OutputString(data=_log)

    def log_node_slab_connectivity(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Node-Slab Connectivity Table  ------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '',
                tabulate(self.input_data['NodeSlabConn'],
                         headers='keys',
                         tablefmt='pipe',
                         floatfmt=".2f"), '', '']
        self.logs.outputStrings['NodeSlabConnTable'] = OutputString(data=_log)

    def process_node_slab_connectivity(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Process node-slab connectivity  ----------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        _df_input = self.input_data['NodeSlabConn'][self.input_data['NodeSlabConn'].Frame_Trans_Loc.notnull()]
        _df_input = _df_input.fillna(value={'SlabConΔ1': '', 'SlabConΔ2': ''})
        _df_input_nodes_dict = self.input_data['Nodes'].to_dict('index')
        _df = pd.DataFrame(columns=['NodeID'])
        _slabs_dict = self.input_data['Slabs'].to_dict('index')

        # import misc info
        _building_type = self.misc['building_type']
        _γσκυρ = self.misc['γσκυρ']
        _gπλ = self.misc['gπλ']
        _qπλ = self.misc['qπλ']
        _hf = self.misc['hf']
        _gτοιχ_in = self.misc['gτοιχ_in']
        _gτοιχ_out = self.misc['gτοιχ_out']
        _hτοιχ = self.misc['hτοιχ']
        _b_beam_y = self.misc['b_beam_y']
        _h_beam_y = self.misc['h_beam_y']
        _L_beam_y = self.misc['L_beam_y']
        _h_storey = self.misc['h_storey']

        _row_data = dict
        # *****************************************************************
        # **** Φορτία στους κόμβους που συνδέονται με εγκάρσιες δοκούς ****
        # *****************************************************************
        for index, row in _df_input.iterrows():
            row_in_nodes_table = _df_input_nodes_dict[index]
            # import from excel table
            _NodeID = index
            _storey = row_in_nodes_table['Storey']
            _pointEtabs = row_in_nodes_table['Point_Etabs']
            _frame_trans_loc = row['Frame_Trans_Loc']
            _slabConnΔ1 = row['SlabConΔ1']
            _slabConnΔ2 = row['SlabConΔ2']

            _col_geom_name = ''
             # ********* Ίδιο βάρος στύλου ***********************
            if _storey != self.stories_total:
                i = self.input_data['Elements'].query(
                    '(NodeI==@_NodeID) & '
                    '((ElementType=="COLUMN_EX") | (ElementType=="COLUMN_IN")'
                    ' | (ElementType=="WALL"))').first_valid_index()
                if i is not None:
                    _col_geom_name = self.input_data['Elements']['SectionName'][i]
                    _b_col, _h_col = self.get_element_b_h(_col_geom_name)
            else:
                _col_geom_name = ''
                _b_col, _h_col = 0., 0.

            # ********* Ίδιο βάρος τοιχοπληρώσεων ****************
            if _frame_trans_loc == 'IN':
                _gτοιχ = _gτοιχ_in
            else:
                _gτοιχ = _gτοιχ_out

            # ********* Φορτία εγκάρσιων δοκών *******************
            if _slabConnΔ1 != '':
                _gΔ1πλ = phd_beam_loads_from_slabs(_slabs_dict, _slabConnΔ1, _gπλ)
                _qΔ1 = phd_beam_loads_from_slabs(_slabs_dict, _slabConnΔ1, _qπλ)
                _gΔ1ΙΒ = _γσκυρ * _b_beam_y * (_h_beam_y - _hf)
                if _storey != self.stories_total:
                    _gΔ1τοιχ = _gτοιχ * _hτοιχ
                else:
                    _gΔ1τοιχ = 0.0
                _gΔ1ολ = _gΔ1πλ + _gΔ1ΙΒ + _gΔ1τοιχ
            else:
                _gΔ1πλ = 0.
                _qΔ1 = 0.
                _gΔ1ΙΒ = 0.
                _gΔ1τοιχ = 0.
                _gΔ1ολ = 0.

            if _slabConnΔ2 != '':
                _gΔ2πλ = phd_beam_loads_from_slabs(_slabs_dict, _slabConnΔ2, _gπλ)
                _qΔ2 = phd_beam_loads_from_slabs(_slabs_dict, _slabConnΔ2, _qπλ)
                _gΔ2ΙΒ = _γσκυρ * _b_beam_y * (_h_beam_y - _hf)
                if _storey != self.stories_total:
                    _gΔ2τοιχ = _gτοιχ * _hτοιχ
                else:
                    _gΔ2τοιχ = 0.0
                _gΔ2ολ = _gΔ2πλ + _gΔ2ΙΒ + _gΔ2τοιχ
            else:
                _gΔ2πλ = 0.
                _qΔ2 = 0.
                _gΔ2ΙΒ = 0.
                _gΔ2τοιχ = 0.
                _gΔ2ολ = 0.

            _G_col = _γσκυρ * _b_col * _h_col * _h_storey

            _G = (_gΔ1ολ + _gΔ2ολ) * _L_beam_y / 2.0 + _G_col
            _Q = (_qΔ1 + _qΔ2) * _L_beam_y / 2.0

            _row_data = {'NodeID': _NodeID,
                          'Storey': _storey,
                          'Point_Etabs': _pointEtabs,
                          'ColGeom': _col_geom_name,
                          'gπλ': _gπλ,
                          'qπλ': _qπλ,
                          'hf': _hf,
                          'gτοιχ': _gτοιχ,
                          'hτοιχ': _hτοιχ,
                          'b_col': _b_col,
                          'h_col': _h_col,
                          'b_beam_y': _b_beam_y,
                          'h_beam_y': _h_beam_y,
                          'gΔ1πλ': _gΔ1πλ,
                          'qΔ1': _qΔ1,
                          'gΔ2πλ': _gΔ2πλ,
                          'qΔ2': _qΔ2,
                          'gΔ1ΙΒ': _gΔ1ΙΒ,
                          'gΔ1τοιχ': _gΔ1τοιχ,
                          'gΔ1': _gΔ1ολ,
                          'gΔ2ΙΒ': _gΔ2ΙΒ,
                          'gΔ2τοιχ': _gΔ2τοιχ,
                          'gΔ2': _gΔ2ολ,
                          'G_col': _G_col,
                          'G': _G,
                          'Q': _Q,
                          }
            _df = _df.append(_row_data, ignore_index=True)

        # *********************************************************************************
        # **** Φορτία στους κόμβους που συνδέονται με τοιχώματα. Ίδιο βάρος τουχωμάτων ****
        # **** Το κάνω ξεχωριστά γιατί δε συνδέονται με εγκάρσιες δοκούς όπως παραπάνω ****
        # *********************************************************************************
        _df_wall_part = self.input_data['Elements'].query('ElementType=="WALL"')
        for index, row in _df_wall_part.iterrows():
            _storey = row['Storey']
            # Ίδιο βάρος τοιχώματος
            if  _storey!= self.stories_total:
                _wall_geom_name = row['SectionName']
                _node_start = row['NodeI']
                _pointEtabs = _df_input_nodes_dict[_node_start]['Point_Etabs']
                _b_wall, _h_wall = self.get_element_b_h(_wall_geom_name)
            else:
                _wall_geom_name = ''
                _b_wall, _h_wall = 0., 0.

            _G_wall = _γσκυρ * _b_wall * _h_wall * _h_storey

            _row_data = {'NodeID': _node_start,
                         'Storey': _storey,
                         'Point_Etabs': _pointEtabs,
                         'ColGeom': _wall_geom_name,
                         'gπλ': 0.0,
                         'qπλ': 0.0,
                         'hf': 0.0,
                         'gτοιχ': 0.0,
                         'hτοιχ': 0.0,
                         'b_col': _b_wall,
                         'h_col': _h_wall,
                         'b_beam_y': 0.0,
                         'h_beam_y': 0.0,
                         'gΔ1πλ': 0.0,
                         'qΔ1': 0.0,
                         'gΔ2πλ': 0.0,
                         'qΔ2': 0.0,
                         'gΔ1ΙΒ': 0.0,
                         'gΔ1τοιχ': 0.0,
                         'gΔ1': 0.0,
                         'gΔ2ΙΒ': 0.0,
                         'gΔ2τοιχ': 0.0,
                         'gΔ2': 0.0,
                         'G_col': _G_wall,
                         'G': _G_wall,
                         'Q': 0.0,
                          }

            _df = _df.append(_row_data, ignore_index=True)

        self.input_data['node_loads'] = _df[_row_data.keys()]

        # *********  Write log  ***********************
        _log.append(tabulate(self.input_data['node_loads'], headers='keys',
                 tablefmt='pipe',
                 floatfmt=".2f"))
        _log.append('')
        _log.append('')
        self.logs.outputStrings['node-slab'] = OutputString(data = _log)

    def log_beam_slab_connectivity(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Beam-Slab Connectivity Table  ------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']
        _log.append(tabulate(self.input_data['BeamSlabConn'], headers='keys',
                 tablefmt='pipe',
                 floatfmt=".2f"))
        _log.append('')
        _log.append('')
        self.logs.outputStrings['BeamSlabConnTable'] = OutputString(data=_log)

    def process_beam_slab_connectivity(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Process beam-slab connectivity  ----------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        _df_input = self.input_data['BeamSlabConn']
        _df = pd.DataFrame(columns=['ElementID', 'SectionName'])

        # import misc info
        _building_type = self.misc['building_type']
        _γσκυρ = self.misc['γσκυρ']
        _gπλ = self.misc['gπλ']
        _qπλ = self.misc['qπλ']
        _qπλ = self.misc['qπλ']
        _hf = self.misc['hf']
        _gτοιχ_in = self.misc['gτοιχ_in']
        _gτοιχ_out = self.misc['gτοιχ_out']
        _hτοιχ = self.misc['hτοιχ']

        _row_data = dict()

        for index, row in _df_input.iterrows():
            _elementID = index
            _geom_name = row['SectionName']
            _storey = row['Storey']
            _lineEtabs = row['Line_Etabs']
            _slabConn = row['SlabConnectivity']
            _frameLoc = row['Frame_Loc']

            # Φορτίο από ίδιο βάρος
            _b, _h = self.get_element_b_h(_geom_name)
            _gΔΙΒ = _γσκυρ * _b * (_h - _hf)

            # Φορτίο από τοιχοπληρώσεις
            if _frameLoc == 'IN':
                _gτοιχ = _gτοιχ_in
            else:
                _gτοιχ = _gτοιχ_out

            if _storey != self.stories_total:
                _gΔτοιχ = _gτοιχ * _hτοιχ
            else:
                _gΔτοιχ = 0.0

            # Φορτία από πλάκες
            _slabs_dict = self.input_data['Slabs'].to_dict('index')
            _gΔπλ = phd_beam_loads_from_slabs(_slabs_dict, _slabConn, _gπλ)
            _qΔ = phd_beam_loads_from_slabs(_slabs_dict, _slabConn, _qπλ)

            # Συνολικό φορτίο
            _gΔολ = _gΔπλ + _gΔΙΒ + _gΔτοιχ

            if row['SectionName'] != 'RIGID':
                _row_data ={'ElementID': _elementID,
                                  'SectionName': _geom_name,
                                  'Storey': _storey,
                                  'Line_Etabs': _lineEtabs,
                                  'SlabConnectivity': _slabConn,
                                  'gπλ': _gπλ,
                                  'qπλ': _qπλ,
                                  'hf': _hf,
                                  'gτοιχ': _gτοιχ,
                                  'hτοιχ': _hτοιχ,
                                  'b': _b,
                                  'h': _h,
                                  'gΔΙΒ': _gΔΙΒ,
                                  'gΔτοιχ': _gΔτοιχ,
                                  'gΔπλ': _gΔπλ,
                                  'gΔ': _gΔολ,
                                  'qΔ': _qΔ}
            else:
                _row_data ={'ElementID': _elementID,
                                  'SectionName': _geom_name,
                                  'Storey': _storey,
                                  'Line_Etabs': _lineEtabs,
                                  'SlabConnectivity': _slabConn,
                                  'gπλ': _gπλ,
                                  'qπλ': _qπλ,
                                  'hf': None,
                                  'gτοιχ': None,
                                  'hτοιχ': None,
                                  'b': None,
                                  'h': None,
                                  'gΔΙΒ': None,
                                  'gΔτοιχ': None,
                                  'gΔπλ': _gΔπλ,
                                  'gΔ': _gΔπλ,
                                  'qΔ': _qΔ}


            _df = _df.append(_row_data, ignore_index=True)
        self.input_data['beam_loads'] = _df[_row_data.keys()]

        # *********  Write log  ***********************
        _log.append(tabulate(self.input_data['beam_loads'], headers='keys',
                             tablefmt='pipe',
                             floatfmt=".2f"))
        _log.append('')
        _log.append('')
        self.logs.outputStrings['beam-slab'] = OutputString(data=_log)


    def process_storey_masses(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  Process beam-slab connectivity  ----------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        # ***********  Masses from nodes  **************************
        m_nodes = []
        _df_input_nodal = self.input_data['node_loads']

        for i in range(0, self.stories_total):
          _g =  _df_input_nodal.query(f'Storey=={i+1}')['G'].sum()
          _q =  _df_input_nodal.query(f'Storey=={i+1}')['Q'].sum()
          m_nodes.append((_g + 0.3 * _q)/9.81)

        # ***********  Masses from beams  **************************
        m_beams = []
        _df_input_nodal = self.input_data['beam_loads']
        for i in range(0, self.stories_total):
            _load = 0.
            beams_in_storey = _df_input_nodal.query(f'Storey=={i+1}')
            for index, row in beams_in_storey.iterrows():
                _element_id = row['ElementID']
                _gbeam = row['gΔ']
                _qbeam = row['qΔ']
                _element_length = self.get_element_length_by_ID(_element_id)
                _load += (_gbeam + 0.3 * _qbeam) * _element_length

            m_beams.append(_load / 9.81)

        # ***********  Total torey masses  *************************
        m_tot = [sum(x) for x in zip(m_nodes, m_beams)]
        self.storey_masses = m_tot
        _data = {'Storey':[i for i in range(1, self.stories_total+1)], 'mass_nodes': m_nodes, 'mass_beams': m_beams, 'mass': m_tot}
        self.input_data['masses'] = pd.DataFrame(data=_data).set_index('Storey')


        # *********  Write log  ***********************
        _log.append(tabulate(self.input_data['masses'], headers='keys',
                             tablefmt='pipe',
                             floatfmt=".2f"))
        _log.append('')
        _log.append('')
        self.logs.outputStrings['masses'] = OutputString(data=_log)

    def process_column_reinforcement(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  ColumnReinforcement Table ----------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']
        _dict = self.input_data['ColumnReinforcement'].fillna({'LongSide': '0Φ0', 'SideHoopLegs':0}).to_dict('index')
        _dict_sections = self.input_data['Sections'].to_dict('index')
        _dict_concrete = self.input_data['Concrete'].to_dict('index')
        for key, value in _dict.items():
            _reinforcementID = key
            _sectionID = key
            _b = _dict_sections[_sectionID]['b']
            _h = _dict_sections[_sectionID]['h']
            _cnom = float(self.misc['cnom'])

            _fc = _dict_concrete['CONC']['fc']
            _Ec = _dict_concrete['CONC']['E']
            _fy = _dict_concrete['CONC']['fy']
            _fyw = _dict_concrete['CONC']['fyw']
            _Es = 200.e6

            _element_id = (self.input_data['Elements'].query(f'ReinforcementID=="{_reinforcementID}"')).index.tolist()[0]
            _element_length = self.get_element_length_by_ID(_element_id)

            _longMain = value["LongMain"]
            _longSide = value["LongSide"]
            _trans = value["Trans"]
            _sideHoopLegs = value["SideHoopLegs"]

            _sectionMaterials = SectionMaterials(fc=_fc, Ec=_Ec, fy=_fy, Es=_Es, fyw=_fyw)

            self.reinforcement_columns_kanepe[f'{_reinforcementID}'] = \
                self.__get_kanepe_Mtheta (b=_b, h=_h, cnom=_cnom, length=_element_length, materials=_sectionMaterials,
                                          long1str=_longMain, long2str=_longMain, transstr=_trans,
                                          longVstr=_longSide, disdiastr='0Φ0', N=0., side_legs=_sideHoopLegs)


        # *********  Write log  ***********************
        _log.append(tabulate(self.input_data['ColumnReinforcement'], headers='keys',
                             tablefmt='pipe',
                             floatfmt=".2f"))
        _log.append('')
        _log.append('')
        self.logs.outputStrings['column_reinforcement'] = OutputString(data=_log)


    def process_beam_reinforcement(self):
        _log = ['---------------------------------------------------------------------------------------------',
                '-------------  BeamReinforcement Table ------------------------------------------------------',
                '---------------------------------------------------------------------------------------------',
                '']

        _dict = self.input_data['BeamReinforcement'].to_dict('index')
        _dict_sections = self.input_data['Sections'].to_dict('index')
        _dict_concrete = self.input_data['Concrete'].to_dict('index')
        for key, value in _dict.items():
            _reinforcementID = key
            _sectionID = str(_reinforcementID[:_reinforcementID.rfind('.')])
            _b = _dict_sections[_sectionID]['b']
            _h = _dict_sections[_sectionID]['h']
            _cnom = float(self.misc['cnom'])

            _fc = _dict_concrete['CONC']['fc']
            _Ec = _dict_concrete['CONC']['E']
            _fy = _dict_concrete['CONC']['fy']
            _fyw = _dict_concrete['CONC']['fyw']
            _Es = 200.e6

            _element_id = (self.input_data['Elements'].query(f'ReinforcementID=="{_reinforcementID}"')).index.tolist()[0]
            _element_length = self.get_element_length_by_ID(_element_id)

            _longTopLeft = value["LongTopLeft"]
            _longBotLeft = value["LongBotLeft"]
            _transLeft = value["TransLeft"]
            _longTopRight = value["LongTopRight"]
            _longBotRight = value["LongBotRight"]
            _transRight = value["TransRight"]

            _sectionMaterials = SectionMaterials(fc=_fc, Ec=_Ec, fy=_fy, Es=_Es, fyw=_fyw)

            self.reinforcement_beams_kanepe[f'{_reinforcementID}.L.neg'] = \
                self.__get_kanepe_Mtheta (b=_b, h=_h, cnom=_cnom, length=_element_length, materials=_sectionMaterials,
                                        long1str=_longTopLeft, long2str=_longBotLeft, transstr=_transLeft)

            self.reinforcement_beams_kanepe[f'{_reinforcementID}.L.pos'] = \
                self.__get_kanepe_Mtheta (b=_b, h=_h, cnom=_cnom, length=_element_length, materials=_sectionMaterials,
                                        long1str=_longBotLeft, long2str=_longTopLeft, transstr=_transLeft)

            self.reinforcement_beams_kanepe[f'{_reinforcementID}.R.neg'] = \
                self.__get_kanepe_Mtheta (b=_b, h=_h, cnom=_cnom, length=_element_length, materials=_sectionMaterials,
                                        long1str=_longTopRight, long2str=_longBotRight, transstr=_transRight)

            self.reinforcement_beams_kanepe[f'{_reinforcementID}.R.pos'] = \
                self.__get_kanepe_Mtheta (b=_b, h=_h, cnom=_cnom, length=_element_length, materials=_sectionMaterials,
                                        long1str=_longBotRight, long2str=_longTopRight, transstr=_transRight)

        # *********  Write log  ***********************
        _log.append(tabulate(self.input_data['BeamReinforcement'], headers='keys',
                             tablefmt='pipe',
                             floatfmt=".2f"))
        _log.append('')
        _log.append('')
        self.logs.outputStrings['beam_reinforcement'] = OutputString(data=_log)

    def __get_kanepe_Mtheta(self, b, h, cnom, length, materials, long1str, long2str, transstr, longVstr='0Φ0', disdiastr='0Φ0', N=0., side_legs=0):

        _reinf = RectangularSectionReinforcement(
            long1=LongReinforcementLayer.from_string(reinf_string=long1str, units_input='mm', units_output='m'),
            long2=LongReinforcementLayer.from_string(reinf_string=long2str, units_input='mm', units_output='m'),
            longV=LongReinforcementLayer.from_string(reinf_string=longVstr, units_input='mm', units_output='m'),
            disdia=LongReinforcementLayer.from_string(reinf_string=disdiastr, units_input='mm', units_output='m'),
            trans=TransReinforcementLayer.from_string(reinf_string=transstr, units_input='mm', units_output='m'),
            cnom=cnom,
            side_hoop_legs= side_legs)

        return RectKanepe(
                    rcs = RectangularConcreteSection(b=b,
                                                     h=h,
                                                     materials = materials,
                                                     reinforcement = _reinf),
                    Ls = length / 2.0,
                    N = N,
                    calc_on_init = False
                )



    def get_element_length_by_ID(self, elementID):
        _nodeI, _nodeJ = self.get_element_endpoints(elementID)

        _nodeI_X = self.input_data['Nodes']['X'][_nodeI]
        _nodeI_Y = self.input_data['Nodes']['Y'][_nodeI]
        _nodeI_Z = self.input_data['Nodes']['Z'][_nodeI]

        _nodeJ_X = self.input_data['Nodes']['X'][_nodeJ]
        _nodeJ_Y = self.input_data['Nodes']['Y'][_nodeJ]
        _nodeJ_Z = self.input_data['Nodes']['Z'][_nodeJ]

        return self.get_element_length(_nodeI_X, _nodeI_Y, _nodeI_Z, _nodeJ_X, _nodeJ_Y, _nodeJ_Z)

    def get_element_endpoints(self, elementID):
        _nodeI = int(self.input_data['Elements']['NodeI'][elementID])
        _nodeJ = int(self.input_data['Elements']['NodeJ'][elementID])
        return _nodeI, _nodeJ


    @staticmethod
    def get_element_length(x_start, y_start, z_start, x_end, y_end, z_end):
        return ((x_end - x_start)**2 + (y_end - y_start)**2 + (z_end - z_start)**2)**0.5


    def get_element_b_h(self, section_name):
        _b = self.input_data['Sections']['b'][section_name]
        _h = self.input_data['Sections']['h'][section_name]
        return _b, _h


    def save_model(self, pathfilename):
        f = gzip.open(pathfilename + '.model', 'wb', compresslevel=9)
        pickle.dump(self, f)
        f.close()


import xlwings as xw
from dataclasses import dataclass
from streng.xwflib.xwf_main import *
from streng.ppp.loads.slabs_to_beams import SlabLoadsToBeams


@dataclass
class slabs_phd:
    typos: str
    Lmax: float
    Lmin: float


# # Πλαισιακοί φορείς
# slabs_frames = {}
# slabs_frames['Π1'] = slabs_phd(typos = '5a',
#                               Lmax = 6.0, Lmin = 3.0)
# slabs_frames['Π2'] = slabs_phd(typos = '6',
#                               Lmax = 4.0, Lmin = 3.0)
# slabs_frames['Π3'] = slabs_frames['Π1']
#
#
# # Δίδυμοι φορείς 59 - 9ώροφα
# slabs_duals9 = {}
# slabs_duals9['Π1'] = slabs_phd(typos = '4',
#                               Lmax = 6.0, Lmin = 3.0)
# slabs_duals9['Π2'] = slabs_phd(typos = '5b',
#                               Lmax = 6.0, Lmin = 3.0)
# slabs_duals9['Π3'] = slabs_duals9['Π1']
#
# slabs_duals9['Π4'] = slabs_phd(typos = '5a',
#                               Lmax = 6.0, Lmin = 3.0)
# slabs_duals9['Π5'] = slabs_phd(typos = '6',
#                               Lmax = 6.0, Lmin = 3.0)
# slabs_duals9['Π6'] = slabs_duals9['Π4']
#
# slabs_duals9['Π7'] = slabs_duals9['Π4']
# slabs_duals9['Π8'] = slabs_duals9['Π5']
# slabs_duals9['Π9'] = slabs_duals9['Π7']
#
#
# # Δίδυμοι φορείς 59 - 4ώροφα
# slabs_duals4 = {}
# slabs_duals4['Π1'] = slabs_phd(typos = '4',
#                               Lmax = 4.5, Lmin = 4.0)
# slabs_duals4['Π2'] = slabs_phd(typos = '5b',
#                               Lmax = 4.5, Lmin = 4.0)
# slabs_duals4['Π3'] = slabs_duals4['Π1']
#
# slabs_duals4['Π4'] = slabs_phd(typos = '5a',
#                               Lmax = 4.5, Lmin = 4.0)
# slabs_duals4['Π5'] = slabs_phd(typos = '6',
#                               Lmax = 4.5, Lmin = 4.0)
# slabs_duals4['Π6'] = slabs_duals4['Π4']
#
# slabs_duals4['Π7'] = slabs_duals4['Π4']
# slabs_duals4['Π8'] = slabs_duals4['Π5']
# slabs_duals4['Π9'] = slabs_duals4['Π7']


@xw.func
def phd_beam_loads_from_slabs(slabs_dict, slab_beam_combo, slab_load):
    _beam_load = 0

    slab_beams = slab_beam_combo.split(',')

    for slab_beam in slab_beams:

        sl_b = slab_beam.split(':')
        sl_name, sl_beam_checked = sl_b[0], int(sl_b[1])

        sltb = SlabLoadsToBeams(l_max = slabs_dict[sl_name]['Lmax'],
                                l_min = slabs_dict[sl_name]['Lmin'],
                                slab_type = slabs_dict[sl_name]['SlabType'],
                                slab_load = slab_load)

        _beam_load += sltb.beams_loads[sl_beam_checked]

    return _beam_load



   





class Beam_uniformP_knownMs:
    def __init__(self, MA, MB, l, p, left_support_name = 'A', rigth_support_name = 'B'):

        """

        Args:
            MA:
            MB:
            l:
            p:
        """
        self.MA = MA
        self.MB = MB
        self.l = l
        self.p = p

        self.left_support_name = left_support_name
        self.rigth_support_name = rigth_support_name

        self.logtext = []

    def calc(self):
        """

        """
        self.MABcenter = 0.5 * (self.MA + self.MB) + self.p * self.l ** 2 / 8
        self.VA = 0.5 * self.p * self.l + (self.MB - self.MA) / self.l
        self.VB = - 0.5 * self.p * self.l + (self.MB - self.MA) / self.l
        self.MABmax = self.MA + self.VA ** 2 / (2 * self.p)
        self.MABmaxr = self.MB + self.VB ** 2 / (2 * self.p)

    def __str__(self):
        self.logtext.append(f'M{self.left_support_name}{self.rigth_support_name}μέσο = {self.MABcenter:.2f}kNm')
        self.logtext.append(f'V{self.left_support_name} = {self.VA:.2f}kN')
        self.logtext.append(f'V{self.rigth_support_name} = {self.VB:.2f}kN')
        self.logtext.append(f'M{self.left_support_name}{self.rigth_support_name}max = {self.MABmax:.2f}kNm')
        self.logtext.append(f'M{self.left_support_name}{self.rigth_support_name}max = {self.MABmaxr:.2f}kNm (υπολογισμένο από δεξιά για επαλήθευση)')

        # print('VA = {0:.2f}kN'.format(VA))
        # print('VB = {0:.2f}kN'.format(VB))
        # print('MABmax = {0:.2f}kNm'.format(MABmax))
        # print('MABmax = {0:.2f}kNm (υπολογισμένο από δεξιά για επαλήθευση)'.format(MABmaxr))
        return '\n'.join(self.logtext)



"""
Concrete elastic deformation functions

    .. uml::

        class elastic_deformation <<(M,#FF7700)>> {
        .. functions ..
        + Ecm(fck)
        + ν(cracked)
        }

"""


import math
# from pyconc.CivilFuncts.Ec2.Ch3.Concrete import Strength

from . import strength


def Ecm(fck):
    """ Modulus of elasticity

    Args:
        fck (float): Characteristic (5%) compressive strength of concrete [MPa]

    Returns:
        float: Given using the expression:

        .. math::
            E_{cm}=22 (\dfrac{f_{cm}}{10})^{0.3}

    """
    return 22 * math.pow((strength.fcm(fck) / 10), 0.3)


def ν(cracked):
    """ Poisson's ratio

    Args:
        cracked (bool): True for cracked, False for uncracked

    Returns:
        float: 0.0 for cracked, 0.2 for uncracked

    """
    if cracked is True:
        return 0.0
    else:
        return 0.2



# -*- coding: utf-8 -*-
import os
import sys
import csv
import numpy
import scipy

from subprocess import STDOUT, check_output, CalledProcessError

from rmtk.vulnerability.common import utils

# DV: this file has too many problems, I give up

def calculate_fragility(capacity_curves, hysteresis, msa, gmrs,gmr_characteristics,damage_model,damping,degradation,number_models_in_DS):
    
    no_damage_states = len(damage_model['damage_states'])
    no_bins = msa['n. bins']
    no_rec_bin = msa['records per bin']
    no_gmrs = no_bins*no_rec_bin
    no_capacity_curves = len(capacity_curves['Sd'])
    
    # dictionary to allocate damage states for all the capacity curves for all the levels of pre-damage
    PDMs, Sds, gmr_info = {},{},{}
    for iDS in range(1,no_damage_states):
        PDMs['iDS'+str(iDS)] = numpy.zeros((no_bins,no_damage_states+1))
        Sds['iDS'+ str(iDS)] = []
        gmr_info['iDS'+ str(iDS)] = []
        
    for icc in range(no_capacity_curves):
        print str((icc+1)*100/no_capacity_curves) + '%'
        # run first Multiple Stripe Analysis            
        run_NLTHA(icc, capacity_curves, hysteresis, gmrs, gmr_characteristics,
                  damage_model, damping, degradation, number_models_in_DS, msa)
        # run second Multiple Stripe Analysis
        PDMs, Sds, gmr_info = run_2MSA(icc,capacity_curves, hysteresis, 
                                       msa,damage_model, gmrs, damping,degradation,
                                       PDMs, Sds, gmr_info,gmr_characteristics)
        
    return PDMs, Sds, gmr_info

def run_NLTHA(icc,capacity_curves,hysteresis,gmrs,gmr_characteristics,damage_model,damping,degradation,number_models_in_DS,msa):
    
    # This function runs a set of NLTHA and writes for each 
    # Damage States a file with the list of records causing the structure to 
    # fall in the given DS. The analyses stop when the number_models_in_DS is 
    # reached for each DS.
    
    no_damage_states = len(damage_model['damage_states'])
    no_gmrs = len(gmrs['time'])
    no_rec_bin = msa['records per bin']
    
    # remove files containing results of previous analyses if they exist
    for iDS in range(0,no_damage_states+1):
        try:
            os.remove('initial_DS%d.csv' %iDS)
        except OSError:
            pass
        
    counter = 0
    # we are running all the gmrs and counter is the variable that records 
    # the count of gmrs already run
    PDM1 = numpy.zeros((no_gmrs,no_damage_states+1))
    store_DS = []
    for iDS in range(0,no_damage_states):
    # in this loop are excluded DS=0 and DS=collapse because they are 
    # not of interest for fragility of damaged buildings
        for igmr in range(counter,no_gmrs):
            limit_states = utils.define_limit_states(capacity_curves,icc,damage_model)
            counter = igmr+1
            target_name = gmrs['name'][igmr]
            available_aftershocks = filter_M_T(target_name,gmr_characteristics)
            if available_aftershocks >= no_rec_bin:
                time, disps = run_time_history_analysis(capacity_curves, hysteresis, icc,gmrs,target_name,1,damping,degradation)
                Sdi = max(numpy.abs(numpy.array(disps)))
                PDM1, ds = utils.allocate_damage(igmr,PDM1,Sdi,limit_states)
                store_DS.append([gmrs['name'][igmr], ds, 1])
            # if maximum number of models in DS is achieved stop
            if sum(PDM1[:,iDS]) >= number_models_in_DS:
                break
            
        # This section works in case not enough records per DS are found
        if sum(PDM1[:,iDS]) < number_models_in_DS and iDS != 0:
            print "WARNING: there are not enough strong ground motions in DS ", str(iDS),", scaling is required"
            for previous_DS in numpy.arange(iDS,0,-1):
                # finding gmrs leading to iDS-1 and applying scaling factor to them
                temp = numpy.array(store_DS)
                list_of_models = temp[scipy.logical_and(temp[:,1] == str(iDS-1),temp[:,2] == str(1))] # names of the ground motion for the previous iDS
                for scaling_factor in numpy.linspace(1.2,3,7):       
                    for ele in list_of_models:
                        limit_states = utils.define_limit_states(capacity_curves,icc,damage_model)
                        gmr_name = ele[0]                        
                        igmr = gmrs['name'].index(gmr_name)
                        # mainshock Mw and Tg are compared to all aftershock Mw and Tg
                        available_aftershocks = filter_M_T(gmr_name,gmr_characteristics)
                        if available_aftershocks < no_rec_bin:
                            index = numpy.where(list_of_models[:,0] == ele[0])
                            list_of_models = numpy.delete(list_of_models, (index), axis=0)
                        else:
                            time, disps = run_time_history_analysis(capacity_curves, hysteresis, icc,gmrs,gmr_name,scaling_factor,damping,degradation)              
                            Sdi = max(numpy.abs(numpy.array(disps)))
                            PDM1, ds = utils.allocate_damage(igmr,PDM1,Sdi,limit_states)
                            store_DS.append([gmr_name, ds, scaling_factor])
                            if ds >= iDS:
                                index = numpy.where(list_of_models[:,0] == ele[0])
                                list_of_models = numpy.delete(list_of_models, (index), axis=0)
                                
                        if sum(PDM1[:,iDS]) >= number_models_in_DS: break
                    if sum(PDM1[:,iDS]) >= number_models_in_DS: break
                if sum(PDM1[:,iDS]) >= number_models_in_DS: break
            if sum(PDM1[:,iDS]) < number_models_in_DS:
                print "WARNING: not enough models in DS%d" %iDS 

        # Store names of records and corresponding DS achieved so far            
        save_DS_list(iDS,store_DS, limit_states)

    
def run_2MSA(icc,capacity_curves,hysteresis,msa,damage_model,gmrs,damping,degradation,PDMs, Sds, gmr_info, gmr_characteristics):
    
    no_damage_states = len(damage_model['damage_states'])
    no_bins = msa['n. bins']
    no_rec_bin = msa['records per bin']
    no_gmrs = no_bins*no_rec_bin
    filter_aftershock = msa['filter']
    Mw_multiplier = gmr_characteristics[-1]
    
    counter = 0
    for iDS in range(1,no_damage_states):
        
        PDM = PDMs['iDS'+str(iDS)]
        # Open files with records name bringing the system to DS = iDS         
        with open('initial_DS%d.csv' %iDS, 'rU') as f:
            gmr_list = [row for row in csv.reader(f)]
            
        for index_gmr1, gmr_name in enumerate(gmr_list):
            print "Initial DS=", iDS
            igmr1 = gmrs['name'].index(gmr_name[0]) # index for 1st gmr within gmrs list
            scl1 = float(gmr_name[1]) #scaling factor for first ground motion
            [M_m, Tg_m, Mw, Tg] = find_M_T(gmr_name[0], gmr_characteristics)
            
            # Read files with records name and corresponding scaling factor for each target IM
            folder_scaled_recs = msa['input folder']
            target_files = os.listdir(folder_scaled_recs)
            for ele in target_files:
            # Remove files that are not csv
                if ele[-3:] != 'csv':
                    target_files.remove(ele)
                    
            for index,ibin in enumerate(target_files):
                if index>=no_bins: break
                print ibin
                
                with open(folder_scaled_recs+'/'+ibin, 'rb') as f:
                    reader = csv.reader(f)
                    newlist = [row for row in reader]
                record_name = [ele[0] for ele in newlist]
                record_scaling = [float(ele[1]) for ele in newlist]
                
                # Run NLTHA with 1st gmr to bring the system to iDS and 2nd gmr corresponding to target IM
                for igmr in range(len(record_name)):
                    limit_states = utils.define_limit_states(capacity_curves,icc,damage_model)
                    target_name = record_name[igmr]
                    scl2 = record_scaling[igmr]
                    [M_a, Tg_a, Mw, Tg] = find_M_T(target_name, gmr_characteristics)
                    if scipy.logical_and(M_a<M_m*Mw_multiplier, Tg_a<Tg_m) or filter_aftershock=='FALSE':
                        igmr2 = gmrs['name'].index(target_name) # index of 2nd gmr within the gmrs list
                        time, disps = run_2time_history_analyses(capacity_curves,hysteresis,icc,gmrs,igmr1,igmr2,scl1,scl2,damping,degradation)
                        Sdi = max(numpy.abs(numpy.array(disps)))
                        # In PDM the number of models in each final damage state corresponding 
                        # to the same IM bin (defined by the variable index) are summed up                     
                        PDM, ds = utils.allocate_damage(index,PDM,Sdi,limit_states)
                        gmr_info['iDS' + str(iDS)].append([igmr2, scl2])
                        Sds['iDS' + str(iDS)].append(Sdi)
                        counter = counter+1
                        print "gmr n.", counter, "max disp. =", Sdi, "DS =", ds
                    if sum(PDM[index,:]) >= no_rec_bin*(index_gmr1+1)*(icc+1): break
                        
        PDMs['iDS' + str(iDS)] = PDM
            
    return PDMs, Sds, gmr_info
    
def save_DS_list(iDS,store_DS,limit_states):
# This function saves in a csv file the names of records leading to each DS

    # open existing file containing names of grms leading to each 
    # DS and continue writing
    out_file = open(os.path.join('initial_DS%d.csv' %iDS), 'a')
    for row in store_DS:
        if row[1] == iDS:
            out_file.write('%s,' %row[0])
            out_file.write('%f\n' %row[2])
    out_file.close()

def run_time_history_analysis(capacity_curves, hysteresis, icc,gmrs,target_name,scaling_factor,damping,degradation):
    
    igmr = gmrs['name'].index(target_name)
    write_gmr(gmrs,igmr, scaling_factor)
    create_SDOF_paramfile(capacity_curves, hysteresis, icc,damping,degradation)
    tcl = 'simplestSDOFgen.tcl'

    try:
        if 'darwin' in sys.platform:
            check_output([os.path.join(bin_directory,
                         "OpenSees.mac"), tcl],
                         cwd=package_directory, stderr=STDOUT)
        elif 'linux' in sys.platform:
            check_output([os.path.join(bin_directory,
                         "OpenSees.u1604"), tcl],
                         cwd=package_directory, stderr=STDOUT)
        else:
            sys.exit('Your OS is not currently support')
    except CalledProcessError as e:
        print(e.output)
        raise

    disps = []
    time = []
    disp_file = open('NodeD.out')
    lines = disp_file.readlines()
    for line in lines:
        values = line.strip().split(' ')
        time.append(float(values[0]))
        disps.append(float(values[1]))
        
    return time, disps
    
def run_2time_history_analyses(capacity_curves, hysteresis,icc,gmrs,igmr1,igmr2,scaling_factor1,scaling_factor2,damping,degradation):
    
    write_double_gmr(gmrs,igmr1,igmr2,scaling_factor1,scaling_factor2)
    
    create_SDOF_paramfile(capacity_curves,hysteresis,icc,damping,degradation)
    os.system('./OpenSees simplestSDOFgen.tcl')
    
    disps = []
    time = []
    disp_file = open('NodeD.out')
    lines = disp_file.readlines()
        
    for line in lines[gmrs['no_points'][igmr1]+2000:]:
        values = line.strip().split(' ')
        time.append(float(values[0]))
        disps.append(float(values[1]))

    return time, disps
    
def write_gmr(gmrs,igmr,scaling_factor):
    
    out_file = open('gmr.txt','w')
    out_file.write(str(gmrs['no_points'][igmr])+'\n')
    out_file.write(str(gmrs['dt'][igmr])+'\n')
    for iacc in range(gmrs['no_points'][igmr]):
        out_file.write(str(gmrs['acc'][igmr][iacc]*scaling_factor)+'\n')
    out_file.close()
    
def write_double_gmr(gmrs,igmr1,igmr2,scaling_factor1,scaling_factor2):
    
    # Harmonise two records
    # change dt in second series if different from previous dt
    if gmrs['dt'][igmr2] != gmrs['dt'][igmr1]:
        time2 = numpy.arange(gmrs['dt'][igmr1],gmrs['time'][igmr2][-1]+gmrs['dt'][igmr1],gmrs['dt'][igmr1])
        gmrs['acc'][igmr2] = numpy.interp(time2,gmrs['time'][igmr2],gmrs['acc'][igmr2])
        gmrs['dt'][igmr2] = gmrs['dt'][igmr1]
        gmrs['time'][igmr2] = time2
        gmrs['no_points'][igmr2] = len(gmrs['acc'][igmr2])
    tot_n_points = gmrs['no_points'][igmr1]+2000+gmrs['no_points'][igmr2]
    
    # first acceleration series
    out_file = open("gmr.txt",'w')
    out_file.write(str(tot_n_points)+'\n')
    out_file.write(str(gmrs['dt'][igmr1])+'\n')
    for iacc in range(gmrs['no_points'][igmr1]):
        out_file.write(str(gmrs['acc'][igmr1][iacc]*scaling_factor1)+'\n')     
    # 20 secs of acc=0 to make the structure come to rest before the aftershock
    for iacc in range(2000):  
        out_file.write(str(0.0)+'\n')
    # second acceleration series
    for iacc in range(gmrs['no_points'][igmr2]):
        out_file.write(str(gmrs['acc'][igmr2][iacc]*scaling_factor2)+'\n')
    out_file.close()
    
def create_SDOF_paramfile(capacity_curves, hysteresis, icc,damping,degradation):

    d_flag = 0
    if degradation:
        d_flag = 1
        
    out_file = open('SDOF_paramfile.tcl','w')
    out_file.write('set material "pinchgeneric" \n')
    f1 = capacity_curves['Sa'][icc][1]*9.81
    f2 = capacity_curves['Sa'][icc][2]*9.81
    f3 = capacity_curves['Sa'][icc][3]*9.81
    f4 = capacity_curves['Sa'][icc][4]*9.81
    d1 = capacity_curves['Sd'][icc][1]
    d2 = capacity_curves['Sd'][icc][2]
    d3 = capacity_curves['Sd'][icc][3]
    d4 = capacity_curves['Sd'][icc][4]
    d5 = capacity_curves['Sd'][icc][4]
    #-capacity_curves['Sd'][icc][1]
    out_file.write('set fi {%f %f %f %f 0.0} \n' %(f1,f2,f3,f4)) 
    out_file.write('set di {%f %f %f %f %f} \n' %(d1,d2,d3,d4,d5))
    out_file.write('set ksi %f \n' %damping)
    out_file.write('set Tperiod "0.0" \n')
    out_file.write('set mx "1.0" \n')
    out_file.write('set degflag "%d" \n' %d_flag)
    out_file.write('set analysistype "dynamic" \n')
    out_file.write('set pinchingtype "%s" \n' %hysteresis['type'])
    
    if hysteresis['type'] == 'user-defined':
        out_file.write('set rDisp %f \n' %hysteresis['rDisp'][0])
        out_file.write('set fForce %f \n' %hysteresis['fForce'][0])
        out_file.write('set uForce %f \n' %hysteresis['uForce'][0])
        gK1, gK2, gK3, gK4, gKLim = hysteresis['gK'][0] 
        gD1, gD2, gD3, gD4, gDLim = hysteresis['gD'][0]
        gF1, gF2, gF3, gF4, gFLim = hysteresis['gF'][0] 
        out_file.write('set gammaK [list %f %f %f %f %f]\n' %(gK1,gK2,gK3,gK4,gKLim))
        out_file.write('set gammaD [list %f %f %f %f %f]\n' %(gD1,gD2,gD3,gD4,gDLim))
        out_file.write('set gammaF [list %f %f %f %f %f]\n' %(gF1,gF2,gF3,gF4,gFLim))
        out_file.write('set gammaE %f \n' %hysteresis['gE'][0])
        out_file.write('set damage "%s"\n' %hysteresis['dmgType'][0])
        
    out_file.close()

def filter_M_T(gmr_name, gmr_characteristics):

    if gmr_characteristics[0] == 'None':
        # gmr_characteristics has not been defined because filtering of aftershocks is not of interest
        return 'Inf'
    else:
        Mw_multiplier = gmr_characteristics[-1]
        [M_m, Tg_m, Mw, Tg] = find_M_T(gmr_name, gmr_characteristics)
        available_aftershocks = len(Mw[scipy.logical_and(Mw<M_m*Mw_multiplier, Tg<Tg_m)])
        return available_aftershocks
    
def find_M_T(gmr_name, gmr_characteristics):
    
    if gmr_characteristics[0] == 'None':
        # gmr_characteristics has not been defined because filtering of aftershocks is not of interest
        return [0, 0, 0, 0]
    else:
        Mw = numpy.array(gmr_characteristics[1])
        Tg = numpy.array(gmr_characteristics[2])
        gmr_list = numpy.array(gmr_characteristics[0])
        M_r = Mw[ gmr_list == gmr_name]
        Tg_r = Tg[ gmr_list == gmr_name]
        
        return [M_r, Tg_r, Mw, Tg]
    



import pandas as pd
import string
import numpy as np
from dataclasses import dataclass, field
from typing import List, Dict
from streng.codes.greek.eak2000.raw.ch2.seismic_action import spectra
from streng.phd.building_models.model2d.slab_loads import *
from streng.phd.building_models.model2d.model_input import Model2d

# from streng.codes.greek.kanepe.assessment.section_properties.rectangular import RectangularKanepe
# from streng.ppp.sections.concrete.typical_sections.rectangular_rc import RectangularConcreteSection
# from streng.ppp.sections.concrete.typical_sections.rectangular_rc import SectionMaterials
# from streng.ppp.sections.concrete.reinforcement.rectangular_reinforcement import RectangularSectionReinforcement
# from streng.ppp.sections.concrete.reinforcement.combos import LongReinforcementLayer, TransReinforcementLayer

# D:\MyBooks\Keimeno\Κεφάλαιο3\Υλικό\FrameAnalyzer\ExcelFiles001\Frames_initial

@dataclass
class EtabsModel2d:
    model: Model2d
    etabs_filename: str = ''
    etabs_file_text: List[str] = field(default_factory=list)

    inelastic_analysis: bool = False

    def etabs_write_file(self):
        self._etabs_write_info()
        self._etabs_write_stories()
        self._etabs_write_diaphragm_names()
        self._etabs_write_grids()
        self._etabs_write_piers()
        self._etabs_write_point_coordinates()
        self._etabs_write_line_connectivities()
        self._etabs_write_point_assigns()
        self._etabs_write_material_properties()
        self._etabs_write_frame_sections()
        if self.inelastic_analysis==True:
            self._etabs_write_hinge_properties()
        self._etabs_write_line_assigns()
        self._etabs_write_static_load_cases()
        self._etabs_write_element_loads()
        self._etabs_write_point_loads()
        self._etabs_write_spectrum_function()
        self._etabs_write_spectrum_cases()
        self._etabs_write_analysis_options()
        if self.inelastic_analysis==True:
            self._etabs_write_pushover_cases()

        self.etabs_file_text.append('  END')
        self.etabs_file_text.append('$ END OF MODEL FILE')

        file = open(self.etabs_filename, 'w')
        file.write('\n'.join(self.etabs_file_text))
        file.close()

    def _etabs_write_info(self):
        self.etabs_file_text.append('$ PROGRAM INFORMATION')
        self.etabs_file_text.append('  PROGRAM  "ETABS"  VERSION "9.7.2"')
        self.etabs_file_text.append('')
        self.etabs_file_text.append('$ CONTROLS')
        self.etabs_file_text.append('  UNITS  "KN"  "M"  ')
        self.etabs_file_text.append('  TITLE1  "university"  ')
        self.etabs_file_text.append('  PREFERENCE  MERGETOL 0.001')
        self.etabs_file_text.append('  RLLF  METHOD "TRIBAREAUBC97"  USEDEFAULTMIN "YES"  ')
        self.etabs_file_text.append('')

    def _etabs_write_analysis_options(self):
        self.etabs_file_text.append('$ ANALYSIS OPTIONS')
        self.etabs_file_text.append('  ACTIVEDOF "UX UZ RY"  ')
        self.etabs_file_text.append(f'  DYNAMICS  MODES {3*self.model.stories_total}  MODETYPE "EIGEN"  TOL 1E-09')
        # self.etabs_file_text.append('  MASSOPTIONS  GRAVITY 9.81  SOURCE "LOADS"  LATERALONLY "YES"    STORYLEVELONLY "YES"  ')
        # self.etabs_file_text.append('  MASSOPTIONS  LOAD "G"  FACTOR 1')
        # self.etabs_file_text.append('  MASSOPTIONS  LOAD "Q"  FACTOR 0.3')
        self.etabs_file_text.append('  MASSOPTIONS  GRAVITY 9.81  SOURCE "MASS"  LATERALONLY "YES"    STORYLEVELONLY "YES"  ')
        self.etabs_file_text.append('')

    def _etabs_write_stories(self):
        self.etabs_file_text.append('$ STORIES - IN SEQUENCE FROM TOP')
        self.etabs_file_text.append(f'  STORY "STORY{self.model.stories_total}"  HEIGHT {self.model.storey_heights[self.model.stories_total - 1]}  MASTERSTORY "Yes"')
        for i in range(self.model.stories_total-1, 0, -1):
            self.etabs_file_text.append(f'  STORY "STORY{i}"  HEIGHT {self.model.storey_heights[i - 1]}  SIMILARTO "STORY{self.model.stories_total}"')
        self.etabs_file_text.append('  STORY "BASE"  ELEV 0')
        self.etabs_file_text.append('')

    def _etabs_write_diaphragm_names(self):
        self.etabs_file_text.append('$ DIAPHRAGM NAMES')
        for i in range(1, self.model.stories_total+1):
            self.etabs_file_text.append(f'  DIAPHRAGM "DIAPH{i}"    TYPE RIGID')
        self.etabs_file_text.append('')

    def _etabs_write_piers(self):
        self.etabs_file_text.append('$ PIER/SPANDREL NAMES')
        self.etabs_file_text.append('  PIERNAME  "P1"')

    def _etabs_write_grids(self):
        alphabet = list(string.ascii_uppercase)
        self.etabs_file_text.append('$ GRIDS')
        self.etabs_file_text.append('  COORDSYSTEM "GLOBAL"  TYPE "CARTESIAN"  BUBBLESIZE 1.25')
        for i, val in enumerate(self.model.x_levels):
            self.etabs_file_text.append(f'  GRID "GLOBAL"  LABEL "{alphabet[i]}"  DIR "X"  COORD {val:.2f}  GRIDTYPE  "PRIMARY"    BUBBLELOC "DEFAULT"  GRIDHIDE "NO"')
        self.etabs_file_text.append('  GRID "GLOBAL"  LABEL "1"  DIR "Y"  COORD 0  GRIDTYPE  "PRIMARY"    BUBBLELOC "SWITCHED"  GRIDHIDE "NO" ')
        self.etabs_file_text.append('')

    def _etabs_write_point_coordinates(self):
        self.etabs_file_text.append('$ POINT COORDINATES')
        for i, val in enumerate(self.model.x_levels):
            self.etabs_file_text.append(f'  POINT "{i+1}"  {val:.2f} 0 ')
        self.etabs_file_text.append('')

    def _etabs_write_line_connectivities(self):
        self.etabs_file_text.append('$ LINE CONNECTIVITIES')


        if self.model.misc['building_type'] == 'frames':
            for i in range(1, len(self.model.x_levels) + 1):
                self.etabs_file_text.append(f'  LINE  "C{i}"  COLUMN  "{i}"  "{i}"  1')
            for i in range(1, len(self.model.x_levels)):
                self.etabs_file_text.append(f'  LINE  "B{i}"  BEAM  "{i}"  "{i+1}"  0')
        else:
            self.etabs_file_text.append(f'  LINE  "C1"  COLUMN  "1"  "1"  1')
            self.etabs_file_text.append(f'  LINE  "C2"  COLUMN  "2"  "2"  1')
            self.etabs_file_text.append(f'  LINE  "C3"  COLUMN  "3"  "3"  1')
            self.etabs_file_text.append(f'  LINE  "C4"  COLUMN  "4"  "4"  1')
            self.etabs_file_text.append(f'  LINE  "C5"  COLUMN  "5"  "5"  1')
            self.etabs_file_text.append(f'  LINE  "C6"  COLUMN  "6"  "6"  1')
            self.etabs_file_text.append(f'  LINE  "C7"  COLUMN  "7"  "7"  1')
            self.etabs_file_text.append(f'  LINE  "C8"  COLUMN  "8"  "8"  1')
            self.etabs_file_text.append(f'  LINE  "C9"  COLUMN  "9"  "9"  1')
            self.etabs_file_text.append(f'  LINE  "C11"  COLUMN  "11"  "11"  1')
            self.etabs_file_text.append(f'  LINE  "C13"  COLUMN  "13"  "13"  1')
            self.etabs_file_text.append(f'  LINE  "B1"  BEAM  "1"  "2"  0')
            self.etabs_file_text.append(f'  LINE  "B2"  BEAM  "2"  "3"  0')
            self.etabs_file_text.append(f'  LINE  "B3"  BEAM  "3"  "4"  0')
            self.etabs_file_text.append(f'  LINE  "B4"  BEAM  "5"  "6"  0')
            self.etabs_file_text.append(f'  LINE  "B5"  BEAM  "6"  "7"  0')
            self.etabs_file_text.append(f'  LINE  "B6"  BEAM  "7"  "8"  0')
            self.etabs_file_text.append(f'  LINE  "B7"  BEAM  "9"  "10"  0')
            self.etabs_file_text.append(f'  LINE  "B8"  BEAM  "10"  "11"  0')
            self.etabs_file_text.append(f'  LINE  "B9"  BEAM  "11"  "12"  0')
            self.etabs_file_text.append(f'  LINE  "B10"  BEAM  "12"  "13"  0')


        self.etabs_file_text.append('')




    def _etabs_write_point_assigns(self):
        self.etabs_file_text.append('$ POINT ASSIGNS')

        masses_dict = self.model.input_data['masses'].to_dict('index')

        # Restraints and diaphragms
        for i in range(1, len(self.model.x_levels) +1):
            self.etabs_file_text.append(f'  POINTASSIGN  "{i}"  "BASE"  RESTRAINT "UX UY UZ RX RY RZ" ')
            for j in range(1, self.model.stories_total + 1):
                self.etabs_file_text.append(f'  POINTASSIGN  "{i}"  "STORY{j}"  DIAPH "DIAPH{j}"  ')

        # Masses
        for i in range(1, self.model.stories_total + 1):
            massX = masses_dict[i]['mass']
            self.etabs_file_text.append(f'  POINTASSIGN  "1"  "STORY{i}"  MASSUXUY {massX:.3f}  ')


        self.etabs_file_text.append('')


    def _etabs_write_material_properties(self):
        self.etabs_file_text.append('$ MATERIAL PROPERTIES')
        # self.etabs_file_text.append('  MATERIAL  "FAKE"  M 0  W 0  TYPE "ISOTROPIC"  E 29000000  U 0.2  A 0')
        # self.etabs_file_text.append('  MATERIAL  "FAKE"  DESIGNTYPE "CONCRETE"  FY 400000  FC 20000  FYS 400000')

        ds = self.model.input_data['Concrete'].to_dict('index')
        for key, value in ds.items():
            _concID = key
            _fc = value['fc']
            _E= value['E']
            _U= value['U']
            _A= value['A']
            _fy= value['fy']
            _fyw= value['fyw']
            self.etabs_file_text.append(f'  MATERIAL  "{_concID}"  M 0  W 0  TYPE "ISOTROPIC"  E {_E}  U {_U}  A {_A}')
            self.etabs_file_text.append(f'  MATERIAL  "{_concID}"  DESIGNTYPE "CONCRETE"  FY {_fy}  FC {_fc}  FYS {_fyw}')
        self.etabs_file_text.append('')

    def _etabs_write_frame_sections(self):
        self.etabs_file_text.append('$ FRAME SECTIONS')
        ds = self.model.input_data['Sections']
        _dict_modFactors = self.model.input_data['ModFactors'].to_dict('index')

        for index, row in ds.iterrows():
            _section_name = index
            _element_type = row['ElementType']
            _b = row['b']
            _h = row['h']
            _hf = self.model.misc['hf']
            _beff = row['beff']
            _frame_loc = row['Frame_Loc']
            _frame = row['Frame']

            mf_area = _dict_modFactors[_element_type]['Area']
            mf_mom = _dict_modFactors[_element_type]['Moment']
            mf_shear = _dict_modFactors[_element_type]['Shear']
            mf_tor = _dict_modFactors[_element_type]['Torsional']

            if _element_type == 'BEAM':
                self.etabs_file_text.append(f'  FRAMESECTION  "{_section_name}"  MATERIAL "CONC"  SHAPE "Tee"  D {_h}  B {_beff}  TF {_hf}  TW {_b}')
            elif 'COLUMN' in _element_type or _element_type == 'WALL' or _element_type == 'RIGID':
                self.etabs_file_text.append(f'  FRAMESECTION  "{_section_name}"  MATERIAL "CONC"  SHAPE "Rectangular"  D {_h}  B {_b}')

            self.etabs_file_text.append(f'  FRAMESECTION  "{_section_name}"  AMOD {mf_area:.2f}  A2MOD {mf_shear:.2f}  A3MOD {mf_shear:.2f}  JMOD {mf_tor:.2f}  I2MOD {mf_mom:.2f}  I3MOD {mf_mom:.2f}  MMOD 0  WMOD 0')

        self.etabs_file_text.append('')

    def _etabs_write_line_assigns(self):
        self.etabs_file_text.append('$ LINE ASSIGNS')
        _dict = self.model.input_data['Elements'].to_dict('index')
        for row in _dict:
            _line = _dict[row]['Line_Etabs']
            _element_type = _dict[row]['ElementType']
            _storey = _dict[row]['Storey']
            _geometry_name = _dict[row]['SectionName']
            if _element_type == 'BEAM':
                self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  SECTION "{_geometry_name}"  ANG  0  MAXSTASPC 0.5  LENGTHOFFI 0  LENGTHOFFJ 0  RIGIDZONE 1  CARDINALPT 8    MESH "POINTSANDLINES"')
                if self.inelastic_analysis==True:
                    self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  HINGE "{_geometry_name}.{_line}.L"  RDISTANCE 0')
                    self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  HINGE "{_geometry_name}.{_line}.R"  RDISTANCE 1')
            elif 'COLUMN' in _element_type:
                self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  SECTION "{_geometry_name}"  ANG  0  MINNUMSTA 3  LENGTHOFFI 0  LENGTHOFFJ 0  RIGIDZONE 1  MESH "POINTSANDLINES"')
                if self.inelastic_analysis==True:
                    self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  HINGE "{_geometry_name}"  RDISTANCE 0')
                    self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  HINGE "{_geometry_name}"  RDISTANCE 1')
            elif _element_type == 'WALL':
                self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  SECTION "{_geometry_name}"  ANG  0  MINNUMSTA 3  LENGTHOFFI 0  LENGTHOFFJ 0  RIGIDZONE 1  MESH "POINTSANDLINES"    PIER "P1"')
            elif _element_type == 'RIGID':
                self.etabs_file_text.append(f'  LINEASSIGN  "{_line}"  "STORY{_storey}"  SECTION "{_geometry_name}"  ANG  0  MAXSTASPC 0.5  LENGTHOFFI 0  LENGTHOFFJ 0  RIGIDZONE 1  CARDINALPT 8    MESH "POINTSANDLINES"')
        self.etabs_file_text.append('')

    def _etabs_write_static_load_cases(self):
        self.etabs_file_text.append('$ STATIC LOADS')
        self.etabs_file_text.append('  LOADCASE "G"  TYPE  "DEAD"  SELFWEIGHT  0')
        self.etabs_file_text.append('  LOADCASE "Q"  TYPE  "LIVE"  SELFWEIGHT  0')
        self.etabs_file_text.append('  LOADCASE "EXSTAT"  TYPE  "QUAKE"  SELFWEIGHT  0')
        self.etabs_file_text.append('  LOADCASE "EYSTAT"  TYPE  "QUAKE"  SELFWEIGHT  0')
        self.etabs_file_text.append('  LOADCASE "EPUSH"  TYPE  "QUAKE"  SELFWEIGHT  0')
        self.etabs_file_text.append('')

    def _etabs_write_element_loads(self):
        self.etabs_file_text.append('$ LINE OBJECT LOADS')
        ds = self.model.input_data['beam_loads']
        for index, row in ds.iterrows():
            _line_etabs = row['Line_Etabs']
            _storey = row['Storey']
            _g = row['gΔ']
            _q = row['qΔ']
            self.etabs_file_text.append(f'  LINELOAD  "{_line_etabs}"  "STORY{_storey:.0f}"  TYPE "UNIFF"  DIR "GRAV"  LC "G"  FVAL {_g:.2f}')
            self.etabs_file_text.append(f'  LINELOAD  "{_line_etabs}"  "STORY{_storey:.0f}"  TYPE "UNIFF"  DIR "GRAV"  LC "Q"  FVAL {_q:.2f}')
        self.etabs_file_text.append('')

    def _etabs_write_point_loads(self):
        self.etabs_file_text.append('$ POINT OBJECT LOADS')
        ds = self.model.input_data['node_loads'] # .query('Storey>=1' )
        for index, row in ds.iterrows():
            _storey = row['Storey']
            _point_etabs = row['Point_Etabs']
            _G = row['G']
            _Q = row['Q']
            self.etabs_file_text.append(f'  POINTLOAD  "{_point_etabs:.0f}"  "STORY{_storey:.0f}"  TYPE "FORCE"  LC "G"    FZ -{_G:.2f}')
            self.etabs_file_text.append(f'  POINTLOAD  "{_point_etabs:.0f}"  "STORY{_storey:.0f}"  TYPE "FORCE"  LC "Q"    FZ -{_Q:.2f}')

        for i in range(self.model.stories_total):
            self.etabs_file_text.append(f'  POINTLOAD  "1"  "STORY{i+1}"  TYPE "FORCE"  LC "EPUSH"  FX {self.model.seismic_static_forces["EPUSHmod"][i]:.3f}')
            self.etabs_file_text.append(f'  POINTLOAD  "1"  "STORY{i+1}"  TYPE "FORCE"  LC "EXSTAT"  FX {self.model.seismic_static_forces["EXSTAT"][i]:.3f}')

        self.etabs_file_text.append('')

    def _etabs_write_spectrum_function(self):
        self.etabs_file_text.append('$ FUNCTIONS')
        self.etabs_file_text.append('  FUNCTION "EAKSPEC"  FUNCTYPE "SPECTRUM"  DAMPRATIO 0.05  SPECTYPE "USER"  ')
        _eak_ground = self.model.misc['EAK_ground']
        _eak_α = self.model.misc['EAK_α']
        x = np.linspace(1e-10, 2, 201)
        y = spectra.Φd(T=x,
                       α=_eak_α*9.81,
                       γI=1.0,
                       T1=spectra.T1(_eak_ground),
                       T2=spectra.T2(_eak_ground),
                       q=3.5,
                       η=1.0,
                       θ=1.0,
                       β0=2.5)
        for i in range(0, len(x)):
            self.etabs_file_text.append(f'  FUNCTION "EAKSPEC"  TIMEVAL "{x[i]:.3f}  {y[i]:.3f}"')
        self.etabs_file_text.append('')

    def _etabs_write_hinge_properties(self):
        self.etabs_file_text.append('$ FRAME HINGE PROPERTIES')
        _dict_beams = self.model.input_data['BeamReinforcement'].to_dict('index')
        for key, value in _dict_beams.items():
            _reinforcementID = key
            self.etabs_file_text.append(self._make_hinge_text_beams(f'{_reinforcementID}'))
        self.etabs_file_text.append('')

        _dict_columns = self.model.input_data['ColumnReinforcement'].to_dict('index')
        for key, value in _dict_columns.items():
            _reinforcementID = key
            self.etabs_file_text.append(self._make_hinge_text_columns(f'{_reinforcementID}'))
        self.etabs_file_text.append('')


    def _make_hinge_text_beams(self, hingeID):
        txt = []

        self.model.reinforcement_beams_kanepe[f'{hingeID}.L.neg'].calculate()
        self.model.reinforcement_beams_kanepe[f'{hingeID}.L.pos'].calculate()
        self.model.reinforcement_beams_kanepe[f'{hingeID}.R.neg'].calculate()
        self.model.reinforcement_beams_kanepe[f'{hingeID}.R.pos'].calculate()


        txt.append(f'  HINGE "{hingeID}.L"  {self.model.reinforcement_beams_kanepe[f"{hingeID}.L.neg"].get_etabs9_7_2_M3(negative=True)}')
        txt.append(f'  HINGE "{hingeID}.L"  {self.model.reinforcement_beams_kanepe[f"{hingeID}.L.pos"].get_etabs9_7_2_M3(negative=False)}')
        txt.append(f'  HINGE "{hingeID}.L"  TYPE "M3"  MOMENTSFP 1  ROTATIONSFP 1')
        txt.append(f'  HINGE "{hingeID}.L"  TYPE "M3"  MOMENTSFN 1  ROTATIONSFN 1')
        txt.append(f'  HINGE "{hingeID}.L"  TYPE "M3"  -IO -2  -LS -4  -CP -6')
        txt.append(f'  HINGE "{hingeID}.L"  TYPE "M3"  IO 2  LS 4  CP 6')

        txt.append(f'  HINGE "{hingeID}.R"  {self.model.reinforcement_beams_kanepe[f"{hingeID}.R.neg"].get_etabs9_7_2_M3(negative=True)}')
        txt.append(f'  HINGE "{hingeID}.R"  {self.model.reinforcement_beams_kanepe[f"{hingeID}.R.pos"].get_etabs9_7_2_M3(negative=False)}')
        txt.append(f'  HINGE "{hingeID}.R"  TYPE "M3"  MOMENTSFP 1  ROTATIONSFP 1')
        txt.append(f'  HINGE "{hingeID}.R"  TYPE "M3"  MOMENTSFN 1  ROTATIONSFN 1')
        txt.append(f'  HINGE "{hingeID}.R"  TYPE "M3"  -IO -2  -LS -4  -CP -6')
        txt.append(f'  HINGE "{hingeID}.R"  TYPE "M3"  IO 2  LS 4  CP 6')


        return '\n'.join(txt)

    def _make_hinge_text_columns(self, hingeID):
        txt = []

        self.model.reinforcement_columns_kanepe[f'{hingeID}'].calculate()

        txt.append(f'  HINGE "{hingeID}"  {self.model.reinforcement_columns_kanepe[f"{hingeID}"].get_etabs9_7_2_M3(negative=True)}')
        txt.append(f'  HINGE "{hingeID}"  {self.model.reinforcement_columns_kanepe[f"{hingeID}"].get_etabs9_7_2_M3(negative=False)}')
        txt.append(f'  HINGE "{hingeID}"  TYPE "M3"  MOMENTSFP 1  ROTATIONSFP 1')
        txt.append(f'  HINGE "{hingeID}"  TYPE "M3"  MOMENTSFN 1  ROTATIONSFN 1')
        txt.append(f'  HINGE "{hingeID}"  TYPE "M3"  -IO -2  -LS -4  -CP -6')
        txt.append(f'  HINGE "{hingeID}"  TYPE "M3"  IO 2  LS 4  CP 6')


        return '\n'.join(txt)


    def _etabs_write_spectrum_cases(self):
        self.etabs_file_text.append('$ RESPONSE SPECTRUM CASES')
        self.etabs_file_text.append('  RSCASE "EX"  DAMP 0.05  DIRCOMBO "SRSS"  ANG 0')
        self.etabs_file_text.append('  RSCASE "EX"  FUNC1 "EAKSPEC"  SF1 1')
        self.etabs_file_text.append('  RSCASE "EX"  MODECOMBO "CQC"')

        self.etabs_file_text.append('')

    def _etabs_write_pushover_cases(self):
        self.etabs_file_text.append('$ RESPONSE SPECTRUM CASES')
        self.etabs_file_text.append('  STATICNLCASE "PUSHGRAV"  CONTROL "FORCE"  STORY "STORY4"  POINT "1"  DOF "UX"  ')
        self.etabs_file_text.append('  STATICNLCASE "PUSHGRAV"  GEONONLIN "NONE"  ')
        self.etabs_file_text.append('  STATICNLCASE "PUSHGRAV"  MINSTEPS 1  MAXNULLSTEPS 50  MAXTOTALSTEPS 200  MAXITER 10')
        self.etabs_file_text.append('  STATICNLCASE "PUSHGRAV"  ITERTOL 0.0001  EVENTTOL 0.01')
        self.etabs_file_text.append('  STATICNLCASE "PUSHGRAV"  LOADTYPE "STATIC"  LOAD "G"  SCALEFACTOR 1')
        self.etabs_file_text.append('  STATICNLCASE "PUSHGRAV"  LOADTYPE "STATIC"  LOAD "Q"  SCALEFACTOR 0.3')
        self.etabs_file_text.append('  STATICNLCASE "PUSHGRAV"  ACTIVEGROUP "ALL"  ')
        self.etabs_file_text.append('  STATICNLCASE "PUSHOVER"  CONTROL "CONJUGATEDISP"  TARGETDISP 0.54  STORY "STORY4"  POINT "1"  DOF "UX" ')
        self.etabs_file_text.append('  STATICNLCASE "PUSHOVER"  STARTFROM "PUSHGRAV"  GEONONLIN "NONE" ')
        self.etabs_file_text.append('  STATICNLCASE "PUSHOVER"  MINSTEPS 100  MAXNULLSTEPS 500  MAXTOTALSTEPS 2000  MAXITER 50')
        self.etabs_file_text.append('  STATICNLCASE "PUSHOVER"  ITERTOL 0.0001  EVENTTOL 0.001')
        self.etabs_file_text.append('  STATICNLCASE "PUSHOVER"  LOADTYPE "STATIC"  LOAD "EPUSH"  SCALEFACTOR 1')
        self.etabs_file_text.append('  STATICNLCASE "PUSHOVER"  ACTIVEGROUP "ALL"  ')



        self.etabs_file_text.append('')








import pandas as pd
import string
import numpy as np
from dataclasses import dataclass, field
from typing import List, Dict
import pickle
import gzip

from streng.codes.greek.eak2000.raw.ch2.seismic_action import spectra
from streng.phd.building_models.model2d.slab_loads import *
from streng.phd.building_models.model2d.model_input import Model2d
import matplotlib.pyplot as plt

from streng.ppp.sections.geometry.tee import TeeSectionGeometry
from streng.ppp.sections.geometry.rectangular import RectangularSectionGeometry
from streng.ppp.opensees.pre.elements.modified_ibarra_krawinkler import ModifiedIbarraKrawinkler

from openseespy import opensees
import math

def ReadRecord(inFilename, outFilename):
    dt = 0.0
    npts = 0

    # Open the input file and catch the error if it can't be read
    inFileID = open(inFilename, 'r')

    # Open output file for writing
    outFileID = open(outFilename, 'w')

    # Flag indicating dt is found and that ground motion
    # values should be read -- ASSUMES dt is on last line
    # of header!!!
    flag = 0

    # Look at each line in the file
    for line in inFileID:
        if line == '\n':
            # Blank line --> do nothing
            continue
        elif flag == 1:
            # Echo ground motion values to output file
            outFileID.write(line)
        else:
            # Search header lines for dt
            words = line.split()
            lengthLine = len(words)

            if lengthLine >= 4:

                if words[0] == 'NPTS=':
                    # old SMD format
                    for word in words:
                        if word != '':
                            # Read in the time step
                            if flag == 1:
                                dt = float(word)
                                break

                            if flag == 2:
                                npts = int(word.strip(','))
                                flag = 0

                            # Find the desired token and set the flag
                            if word == 'DT=' or word == 'dt':
                                flag = 1

                            if word == 'NPTS=':
                                flag = 2


                elif words[-1] == 'DT':
                    # new NGA format
                    count = 0
                    for word in words:
                        if word != '':
                            if count == 0:
                                npts = int(word)
                            elif count == 1:
                                dt = float(word)
                            elif word == 'DT':
                                flag = 1
                                break

                            count += 1

    inFileID.close()
    outFileID.close()

    return dt, npts


@dataclass
class OpenSeesResults:
    elastic_periods: list = field(init=False, repr=False, default_factory=list)
    elastic_eigenvectors: list = field(init=False, repr=False, default_factory=list)
    elastic_gravity_element_forces: Dict[int, list] = field(init=False, repr=False, default_factory=dict)

    model_mIK_periods: list = field(init=False, repr=False, default_factory=list)
    model_mIK_eigenvectors: list = field(init=False, repr=False, default_factory=list)

    pushover_element_deformation: Dict[int, list] = field(init=False, repr=False, default_factory=dict)
    pushover_element_force: Dict[int, list] = field(init=False, repr=False, default_factory=dict)


@dataclass
class OpenSeesModel2d:
    model: Model2d

    tcl_make: bool = False
    tcl_filename: str = ''
    tcl_file_text: List[str] = field(default_factory=list)

    include_pdelta: bool = False

    inelastic_analysis: bool = False
    inelastic_analysis_type: str = '' # 'pushover' or 'time_history'

    # time_history: bool = False
    # pushover: bool = False

    etabs_pushocer_curve_check: str = ''

    results: OpenSeesResults = field(init=False, repr=False, default_factory=OpenSeesResults)


    def os_do_it(self):
        # self._os_start()
        # self._os_nodes()
        # self._os_node_fixes()
        # self._os_node_masses()
        # self._os_node_equal_displ()
        # self._os_elements()
        #
        # # if self.inelastic_analysis == True:
        # #     self._os_zero_length_elements()
        #
        #
        #
        # self.inelastic_analysis = True
        # if self.inelastic_analysis == True:
        #     self._os_nodes_extra()
        #
        # self.results.periods_elastic, self.results.eigenvectors_elastic = self._os_eigen_analysis()

        # self._os_gravity_loads()
        #
        # self._os_run_elastic_analysis()
        #
        # if self.pushover == True:
        #     self._os_run_pushover()
        #
        # if self.time_history == True:
        #     self._os_run_time_history()


        file = open(self.tcl_filename, 'w')
        file.write('\n'.join(self.tcl_file_text))
        file.close()


    def os_run_elastic(self):
        self._os_start()
        self._os_nodes()
        self._os_node_fixes()
        self._os_node_masses()
        self._os_node_equal_displ()
        self._os_elements(include_zero_length=False)
        self.results.elastic_periods, self.results.elastic_eigenvectors = self._os_eigen_analysis()
        self._os_gravity_loads()
        self._os_run_gravity_elastic()


    def os_run_pushover(self):
        self._os_start()
        self._os_nodes()
        self._os_nodes_extra()
        self._os_node_fixes()
        self._os_node_masses()
        self._os_node_equal_displ()
        self._os_elements(include_zero_length=True)
        self._os_zero_length_elements()
        self.results.model_mIK_periods, self.results.model_mIK_eigenvectors = self._os_eigen_analysis()
        self._os_run_push_grav()
        self._os_run_pushover()


    def os_run_time_history(self):
        self._os_start()
        self._os_nodes()
        self._os_nodes_extra()
        self._os_node_fixes()
        self._os_node_masses()
        self._os_node_equal_displ()
        self._os_elements(include_zero_length=True)
        self._os_zero_length_elements()
        self.results.model_mIK_periods, self.results.model_mIK_eigenvectors = self._os_eigen_analysis()
        self._os_run_push_grav()
        self.__os_run_time_history()



    def _os_start(self):
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('#')
        self.tcl_file_text.append('# M O D E L  D O M A I N  1  (3DOF)')
        self.tcl_file_text.append('#')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('')

        opensees.wipe()
        self.tcl_file_text.append('wipe')
        opensees.model('Basic', '-ndm', 2)
        self.tcl_file_text.append('model BasicBuilder -ndm 2 -ndf 3')
        self.tcl_file_text.append('source rotSpring2DModIKModel.tcl')
        self.tcl_file_text.append('')

        if self.include_pdelta == False:
            opensees.geomTransf('Linear', 1)
            self.tcl_file_text.append('geomTransf Linear 1')
        else:
            opensees.geomTransf('PDelta', 1)
            self.tcl_file_text.append('geomTransf PDelta 1')


    def _os_nodes(self):
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('# N O D E S')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# node $NodeTag $XCoord $Ycoord')
        self.tcl_file_text.append('')

        for index, row in self.model.input_data['Nodes'].iterrows():
            # print(f'{index} {float(row["X"])} {row["Z"]}')
            opensees.node(index, float(row['X']), row['Z'])
            self.tcl_file_text.append(f'node{index:7}{float(row["X"]):13.2f}{float(row["Z"]):13.2f}')


    def _os_nodes_extra(self):
        self.tcl_file_text.append('# -----------------------------------------------')
        self.tcl_file_text.append('# Nodes for zero length elements')
        self.tcl_file_text.append('# -----------------------------------------------')

        dfElements = self.model.input_data['Elements']
        nodes_dict = self.model.input_data['Nodes'].to_dict('index')

        for index, row in dfElements.iterrows():
            _nodeI, _nodeJ = self.model.get_element_endpoints(index)
            if row['ElementType'] == 'BEAM':
                _extra_nodeI = 10 * _nodeI + 2
                _extra_nodeJ = 10 * _nodeJ + 4

            if 'COLUMN' in str(row['ElementType']):
                _extra_nodeI = 10 * _nodeI + 1
                _extra_nodeJ = 10 * _nodeJ + 3

            _extra_nodeI_X = float(nodes_dict[_nodeI]['X'])
            _extra_nodeI_Y = float(nodes_dict[_nodeI]['Z'])
            _extra_nodeJ_X = float(nodes_dict[_nodeJ]['X'])
            _extra_nodeJ_Y = float(nodes_dict[_nodeJ]['Z'])
            opensees.node(_extra_nodeI, _extra_nodeI_X, _extra_nodeI_Y)
            opensees.node(_extra_nodeJ, _extra_nodeJ_X, _extra_nodeJ_Y)

            self.tcl_file_text.append(f'node{_extra_nodeI:7}{_extra_nodeI_X:13.2f}{_extra_nodeI_Y:13.2f}')
            self.tcl_file_text.append(f'node{_extra_nodeJ:7}{_extra_nodeJ_X:13.2f}{_extra_nodeJ_Y:13.2f}')




    def _os_node_fixes(self):
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('# R E S T R A I N T S')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# fix $NodeTag x-transl y-transl z-rot')
        self.tcl_file_text.append('')
        for index, row in self.model.input_data['Nodes'].iterrows():
            if row['Z'] == 0.0:
                opensees.fix(index, 1, 1, 1)
                self.tcl_file_text.append(f'fix{index:7}{1:5}{1:5}{1:5}')


    def _os_node_masses(self):
        if self.tcl_make == True:
            self.tcl_file_text.append('')
            self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
            self.tcl_file_text.append('# M A S S E S')
            self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
            self.tcl_file_text.append('')
            self.tcl_file_text.append('# Mass Definition : mass $NodeTag $(ndf nodal mass values corresponding to each DOF)')
            self.tcl_file_text.append('')

            masses_dict = self.model.input_data['masses'].to_dict('index')

            for i in range(1, self.model.stories_total+1):
                # node = (self.input_data['Nodes']['Storey'] == i).idxmax()
                node = self.model.master_nodes[i-1]
                massX = masses_dict[i]['mass']
                opensees.mass(int(node), massX, 1.0e-10, 1.0e-10)
                self.tcl_file_text.append(f'mass{node:7}{massX:10.3f}{0:10.1f}{0:10.1f}')


    def _os_node_equal_displ(self):
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('# E Q U A L  C O N S T R A I N T S')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# Equal Constraint/equalDOF Definition : equalDOF $MasterNode $SlaveNode $DOFs')
        self.tcl_file_text.append('')
        dfNodes = self.model.input_data['Nodes']
        for storey_level in self.model.storey_levels[1:]:
            nodes_in_level = dfNodes[dfNodes.Z==storey_level].index.tolist()
            for node in nodes_in_level[1:]:
                opensees.equalDOF(nodes_in_level[0], node, 1)
                self.tcl_file_text.append(f'equalDOF{int(nodes_in_level[0]):8}{int(node):10}{1:7}      ; # StoreyLevel: {storey_level:5.2f}')


    def _os_elements(self, include_zero_length):
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('# E L A S T I C   T I M O S H E N K O   B E A M - C O L U M N   E L E M E N T S')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# element ElasticTimoshenkoBeam $eleTag $iNode $jNode $E $G $A $Iz $Avy $transfTag <-mass $massDens> <-cMass>')
        self.tcl_file_text.append('')

        conc_dict = self.model.input_data['Concrete'].to_dict('index')
        modFactors_dict = self.model.input_data['ModFactors'].to_dict('index')

        E = float(conc_dict['CONC']['E'])
        U = float(conc_dict['CONC']['U'])
        G = E / (2.0*(1.0+U))

        dfElements = self.model.input_data['Elements']
        sections_dict = self.model.input_data['Sections'].to_dict('index')
        for index, row in dfElements.iterrows():
            element_type = row['ElementType']
            section_name = row['SectionName']

            bw, h, beff, frame_location = sections_dict[section_name]['b'],\
                                          sections_dict[section_name]['h'], \
                                          sections_dict[section_name]['beff'], \
                                          sections_dict[section_name]['Frame_Loc']
            hf = self.model.misc['hf']

            A = 0.0
            Iz = 0.0
            Avy = 0.0

            mf_area = modFactors_dict[element_type]['Area']
            mf_mom = modFactors_dict[element_type]['Moment']
            mf_shear = modFactors_dict[element_type]['Shear']
            # mf_tor = modFactors_dict[element_type]['Torsional']

            if element_type == 'BEAM':
                tbeam = TeeSectionGeometry(bw = bw, h = h, beff=beff, hf = hf)
                A = tbeam.area * mf_area
                Iz = tbeam.moment_of_inertia_xx * mf_mom
                Avy = tbeam.shear_area_2 * mf_shear

            elif element_type == 'COLUMN_EX':
                rect = RectangularSectionGeometry(b=bw, h=h)
                A = rect.area * mf_area
                Iz = rect.moment_of_inertia_xx * mf_mom
                Avy = rect.shear_area_2 * mf_shear
            elif element_type == 'COLUMN_IN':
                rect = RectangularSectionGeometry(b=bw, h=h)
                A = rect.area * mf_area
                Iz = rect.moment_of_inertia_xx * mf_mom
                Avy = rect.shear_area_2 * mf_shear
            elif element_type == 'WALL':
                rect = RectangularSectionGeometry(b=bw, h=h)
                A = rect.area * mf_area
                Iz = rect.moment_of_inertia_xx * mf_mom
                Avy = rect.shear_area_2 * mf_shear
            elif element_type == 'RIGID':
                rect = RectangularSectionGeometry(b=bw, h=h)
                A = rect.area * mf_area
                Iz = rect.moment_of_inertia_xx * mf_mom
                Avy = rect.shear_area_2 * mf_shear

            n_MIK = 10
            if include_zero_length == True:
                Izmod = Iz * (n_MIK+1.0) / n_MIK
            else:
                Izmod = Iz

            _nodeI = int(row["NodeI"])
            _nodeJ = int(row["NodeJ"])
            if include_zero_length == True:
                if element_type == 'BEAM':
                    _nodeI = _nodeI * 10 + 2
                    _nodeJ = _nodeJ * 10 + 4
                elif 'COLUMN' in str(element_type):
                    _nodeI = _nodeI * 10 + 1
                    _nodeJ = _nodeJ * 10 + 3
            # if element_type == 'BEAM':
            #     _nodeI = _nodeI * 10 + 2
            #     _nodeJ = _nodeJ * 10 + 4
            # elif 'COLUMN' in str(element_type):
            #     _nodeI = _nodeI * 10 + 1
            #     _nodeJ = _nodeJ * 10 + 3


            opensees.element('ElasticTimoshenkoBeam', int(index), _nodeI, _nodeJ, E, G, A, Izmod, Avy, 1)  # , '-mass', 0., '-lMass')
            self.tcl_file_text.append(f'element ElasticTimoshenkoBeam{index:7}{_nodeI:7}{_nodeJ:7}'
                                      f'{E:13.3e}{G:13.5e}{A:13.5e}{Izmod:13.5e}{Avy:13.5e}'
                                      f'    1   -mass        0')
            # self.tcl_file_text.append(f'element elasticBeamColumn{index:7}{_nodeI:7}{_nodeJ:7}'
            #                           f'{A:13.5e}{E:13.3e}{Izmod:13.5e}     1')


    def _os_zero_length_elements(self):
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('# Z E R O   L E N G T H   E L E M E N T S')
        self.tcl_file_text.append('# --------------------------------------------------------------------------------------------------------------')

        conc_dict = self.model.input_data['Concrete'].to_dict('index')
        modFactors_dict = self.model.input_data['ModFactors'].to_dict('index')


        Ec = float(conc_dict['CONC']['E'])

        n_mIK = 10.

        dfElements = self.model.input_data['Elements']
        sections_dict = self.model.input_data['Sections'].to_dict('index')
        for index, row in dfElements.iterrows():
            element_type = row['ElementType']
            section_name = row['SectionName']
            reinforcement_name = row['ReinforcementID']

            mf_mom = modFactors_dict[element_type]['Moment']

            _nodeI, _nodeJ = self.model.get_element_endpoints(index)

            length = self.model.get_element_length_by_ID(index)

            bw, h, frame_location = sections_dict[section_name]['b'],\
                                   sections_dict[section_name]['h'], \
                                   sections_dict[section_name]['Frame_Loc']
            hf = self.model.misc['hf']

            if element_type == 'BEAM':
                tbeam = TeeSectionGeometry(bw = bw, h = h, beff=bw + 8.0 * hf, hf = hf)
                Iz = tbeam.moment_of_inertia_xx * mf_mom
                _extra_nodeI = 10 * _nodeI + 2
                _extra_nodeJ = 10 * _nodeJ + 4

                McMy = 1.001   # ratio of capping moment to yield moment, Mc / My

                # Left
                hinge_neg = self.model.reinforcement_beams_kanepe[f'{reinforcement_name}.L.neg']
                hinge_pos = self.model.reinforcement_beams_kanepe[f'{reinforcement_name}.L.pos']
                hinge_neg.calculate()
                hinge_pos.calculate()

                print(f'{index}: {reinforcement_name} Left:  My_P={hinge_pos.My}    My_N={-hinge_neg.My}')

                if self.inelastic_analysis_type != 'pushover':
                    beam_mIK_L = ModifiedIbarraKrawinkler(n=n_mIK, E=Ec, I=Iz, L=length,
                                                    My_P=hinge_pos.My, McMy_P=McMy, th_pP=hinge_pos.θpl, th_pcP=0.5*hinge_pos.θpl,
                                                    th_uP=4*hinge_pos.θpl,
                                                    My_N=-hinge_neg.My, McMy_N=McMy, th_pN=hinge_neg.θpl, th_pcN=0.5*hinge_neg.θpl,
                                                    th_uN=4*hinge_neg.θpl)
                else:
                    beam_mIK_L = ModifiedIbarraKrawinkler(n=n_mIK, E=Ec, I=Iz, L=length,
                                                    My_P=hinge_pos.My, McMy_P=McMy, th_pP=hinge_pos.θpl, th_pcP=0.5*hinge_pos.θpl,
                                                    th_uP=4*hinge_pos.θpl,
                                                    My_N=-hinge_pos.My, McMy_N=McMy, th_pN=hinge_pos.θpl, th_pcN=0.5*hinge_pos.θpl,
                                                    th_uN=4*hinge_pos.θpl)


                beam_mIK_L.opensees_peak_py(index * 10 + 1, _nodeI, _extra_nodeI)
                self.tcl_file_text.append(beam_mIK_L.opensees_command_tcl(index * 10 + 1, _nodeI, _extra_nodeI) + f'   # beam {index} {reinforcement_name}.L')
                # print(f'{_nodeI}-{_extra_nodeI}:{reinforcement_name}Left Mypos={hinge_pos.My:.2f}  Myneg={hinge_neg.My:.2f}')

                # Right
                hinge_neg = self.model.reinforcement_beams_kanepe[f'{reinforcement_name}.R.neg']
                hinge_pos = self.model.reinforcement_beams_kanepe[f'{reinforcement_name}.R.pos']
                hinge_neg.calculate()
                hinge_pos.calculate()

                print(f'{index}: {reinforcement_name} Right:  My_P={hinge_pos.My}    My_N={-hinge_neg.My}')


                if self.inelastic_analysis_type != 'pushover':
                    beam_mIK_R = ModifiedIbarraKrawinkler(n=n_mIK, E=Ec, I=Iz, L=length,
                                                    My_P=hinge_pos.My, McMy_P=McMy, th_pP=hinge_pos.θpl, th_pcP=0.5*hinge_pos.θpl,
                                                    th_uP=4*hinge_pos.θpl,
                                                    My_N=-hinge_neg.My, McMy_N=McMy, th_pN=hinge_neg.θpl, th_pcN=0.5*hinge_neg.θpl,
                                                    th_uN=4*hinge_neg.θpl)
                else:
                    beam_mIK_R = ModifiedIbarraKrawinkler(n=n_mIK, E=Ec, I=Iz, L=length,
                                                    My_P=hinge_neg.My, McMy_P=McMy, th_pP=hinge_neg.θpl, th_pcP=0.5*hinge_neg.θpl,
                                                    th_uP=4*hinge_neg.θpl,
                                                    My_N=-hinge_neg.My, McMy_N=McMy, th_pN=hinge_neg.θpl, th_pcN=0.5*hinge_neg.θpl,
                                                    th_uN=4*hinge_neg.θpl)

                beam_mIK_R.opensees_peak_py(index * 10 + 2, _nodeJ , _extra_nodeJ)
                self.tcl_file_text.append(beam_mIK_R.opensees_command_tcl(index * 10 + 2, _nodeJ, _extra_nodeJ) + f'   # beam {index} {reinforcement_name}.R')
                # print(f'{_nodeJ}-{_extra_nodeJ}:{reinforcement_name}Right Mypos={hinge_pos.My:.2f}  Myneg={hinge_neg.My:.2f}')


            if 'COLUMN' in str(row['ElementType']):
                col = RectangularSectionGeometry(b=bw, h=h)
                Iz = col.moment_of_inertia_xx * mf_mom
                _extra_nodeI = 10 * _nodeI + 1
                _extra_nodeJ = 10 * _nodeJ + 3

                hinge = self.model.reinforcement_columns_kanepe[f'{reinforcement_name}']
                hinge.calculate()

                McMy = 1.001   # ratio of capping moment to yield moment, Mc / My

                col_mIK = ModifiedIbarraKrawinkler(n=n_mIK, E=Ec, I=Iz, L=length,
                                                    My_P=hinge.My, McMy_P=McMy, th_pP=hinge.θpl, th_pcP=0.5*hinge.θpl,
                                                    th_uP=4*hinge.θpl,
                                                    My_N=-hinge.My, McMy_N=McMy, th_pN=hinge.θpl, th_pcN=0.5*hinge.θpl,
                                                    th_uN=4*hinge.θpl)

                # Bottom
                col_mIK.opensees_peak_py(index * 10 + 1, _nodeI, _extra_nodeI)
                self.tcl_file_text.append(col_mIK.opensees_command_tcl(index * 10 + 1, _nodeI, _extra_nodeI) + f'   # col {index} {reinforcement_name}.Bot')
                # print(f'{_nodeI}-{_extra_nodeI}:{reinforcement_name} My={hinge.My:.2f}')
                # Top
                col_mIK.opensees_peak_py(index * 10 + 2, _nodeJ, _extra_nodeJ)
                self.tcl_file_text.append(col_mIK.opensees_command_tcl(index * 10 + 2, _extra_nodeJ, _nodeJ) + f'   # col {index} {reinforcement_name}.Top')
                # print(f'{_nodeJ}-{_extra_nodeJ}:{reinforcement_name} My={hinge.My:.2f}')


    def _os_eigen_analysis(self):
        numEigen = self.model.stories_total
        eigenValues = opensees.eigen('-fullGenLapack', numEigen)

        _periods = []
        _eigenvectors = []
        for i in range(0, numEigen):
            lamb = eigenValues[i]
            period = 2 * math.pi / math.sqrt(lamb)
            _periods.append(period)
            print(f'Period {i+1} = {period:.4f}s')

            eigen = []
            for mn in self.model.master_nodes:
                # print(f'eigen {i} mn {mn}')
                eigen.append(opensees.nodeEigenvector(mn, i+1, 1))
            _eigenvectors.append(eigen)


        self.tcl_file_text.append('')
        self.tcl_file_text.append(
            '# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('# E I G E N V A L U S   A N A L Y S I S   -   C A L C U L A T E   P E R I O D S')
        self.tcl_file_text.append(
            '# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('')
        self.tcl_file_text.append(f'eigen -fullGenLapack {numEigen}')
        self.tcl_file_text.append(f'set numModes {numEigen}')
        self.tcl_file_text.append(f'set lambda [eigen -fullGenLapack $numModes]')
        self.tcl_file_text.append(f'puts Periods')
        self.tcl_file_text.append('foreach lam $lambda {')
        self.tcl_file_text.append(f'    puts [expr 6.283185/sqrt($lam)]')
        self.tcl_file_text.append('}')

        return _periods, _eigenvectors

    def _os_gravity_loads(self):
        self.tcl_file_text.append('')
        self.tcl_file_text.append(
            '# --------------------------------------------------------------------------------------------------------------')
        self.tcl_file_text.append('# G R A V I T Y    L O A D S')
        self.tcl_file_text.append(
            '# --------------------------------------------------------------------------------------------------------------')

        self.tcl_file_text.append('timeSeries Linear 1')
        self.tcl_file_text.append('')
        self.tcl_file_text.append('pattern Plain 101 1 {')

        opensees.timeSeries('Linear', 1)
        opensees.pattern('Plain', 101, 1)
        self._os_nodal_loads()
        self._os_element_loads()

        if self.tcl_make == True:
            self.tcl_file_text.append('}')


    def _os_nodal_loads(self):
        ds = self.model.input_data['node_loads'] # .query('Storey>=1' )
        for index, row in ds.iterrows():
            _storey = row['Storey']
            _point_etabs = row['Point_Etabs']
            _node = row['NodeID']
            _G = float(row['G'])
            _Q = float(row['Q'])
            _loadZ = _G+0.3*_Q
            if _loadZ>0:
                opensees.load(_node, 0., -_loadZ, 0.)
                self.tcl_file_text.append(f'    load   {_node}   0.0    {-_loadZ:.2f}   0.0')



    def _os_element_loads(self):
        ds = self.model.input_data['beam_loads']
        for index, row in ds.iterrows():
            _ElementID  = int(row['ElementID'])
            _g = float(row['gΔ'])
            _q = float(row['qΔ'])
            _loadZ = _g + 0.3 * _q
            if _loadZ>0:
                opensees.eleLoad('-ele', _ElementID, '-type', '-beamUniform', -_loadZ)
                if self.tcl_make == True:
                    self.tcl_file_text.append(f'    eleLoad -ele {_ElementID} -type -beamUniform {-_loadZ:.2f}')



    def _os_run_gravity_elastic(self):


        opensees.system('BandGeneral')
        opensees.constraints('Plain')
        opensees.numberer('RCM')

        opensees.algorithm('Linear')
        opensees.integrator('LoadControl', 1.)
        opensees.analysis('Static')
        opensees.analyze(1)


        _df_elements = self.model.input_data['Elements']
        for index, row in _df_elements.iterrows():
            self.results.elastic_gravity_element_forces[index] = opensees.eleResponse(int(index), 'force', 1)
        opensees.wipe('all')



    def _os_run_push_grav(self):
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# ---------------------------------')
        self.tcl_file_text.append('# Create Analysis for Gravity Loads')
        self.tcl_file_text.append('# ---------------------------------')
        self.tcl_file_text.append('')
        self.tcl_file_text.append('constraints Plain;               # Create the constraint handler, a Plain handler is used as homo constraints')
        self.tcl_file_text.append('numberer RCM;                    # Create the DOF numberer, the reverse Cuthill-McKee algorithm')
        self.tcl_file_text.append('system BandGeneral;              # Create the system of equation, a SPD using a band storage scheme')
        self.tcl_file_text.append('test NormDispIncr 1.0e-6 6;      # Create the solution algorithm, a Linear algorithm is created')
        self.tcl_file_text.append('algorithm Newton;')
        self.tcl_file_text.append('integrator LoadControl 0.1;      # Create the integration scheme, the LoadControl scheme using steps of 1.0')
        self.tcl_file_text.append('analyses Static;                 # create the analyses object')
        self.tcl_file_text.append('analyze 10;                      # Perform Gravity Analysis')
        self.tcl_file_text.append('')
        self.tcl_file_text.append('recorder Element -file EleForces.out -ele 1211 1221 1231 1241 forces')
        self.tcl_file_text.append('record')
        self.tcl_file_text.append('exit')

        opensees.constraints('Plain')
        opensees.numberer('RCM')
        opensees.system('BandGeneral')
        opensees.test('NormDispIncr', 1e-6, 6)
        opensees.algorithm('Newton')
        opensees.integrator('LoadControl', 0.1)
        opensees.analysis('Static')
        opensees.analyze(10)

        opensees.loadConst('-time', 0.0)

        print('Some gravity analyses results')
        print(opensees.nodeDisp(106))
        print(opensees.eleForce(1111))


    def _os_run_pushover(self):
        self.tcl_file_text.append('')
        self.tcl_file_text.append('# ---------------------------------')
        self.tcl_file_text.append('# Create Pushover Analysis')
        self.tcl_file_text.append('# ---------------------------------')
        self.tcl_file_text.append('')

        _df_nodes = self.model.input_data['Nodes']
        _df_elements = self.model.input_data['Elements']

        roof_node = max(self.model.master_nodes)

        opensees.recorder('Node', '-file', 'droof.out', '-node', roof_node, '-dof', 1, 'disp')
        opensees.recorder('Node', '-file', 'Vbase.out', '-node', 1011, 1021, 1031, 1041, '-dof', 1, 'reaction')

        # Ακόμα και αν δεν τυπώνω σε αρχείο, πρέπει να τρέχει το recorder για να βγάλει τις λίστες...γιατί όμως;;
        # Δεν είναι ανάγκη να υπάρχουν όλοι οι κόμβοι, φτάνει ένας, ακόμα και αν δεν είναι από αυτούς που ζητώ στη συνέχεια
        opensees.recorder('Node', '-file', 'base_level_nodes.out', '-node', self.model.base_level_nodes[0], '-dof', 1, 'reaction')

        opensees.timeSeries('Linear', 2)
        opensees.pattern('Plain', 200, 2)
        for node in self.model.master_nodes:
            opensees.load(node, 1.0, 0., 0.)

        # opensees.constraints('Plain')
        opensees.numberer('RCM')
        opensees.system('UmfPack')
        opensees.test('NormDispIncr', 1e-6, 2000)
        opensees.algorithm('Newton')
        opensees.integrator('DisplacementControl', roof_node, 1, 0.0001)
        opensees.analysis('Static')

        roof_displacements = [0]
        base_shears = [0]
        for index, row in _df_elements.iterrows():
            self.results.pushover_element_deformation[index * 10 + 1] = []
            self.results.pushover_element_deformation[index * 10 + 2] = []
            self.results.pushover_element_force[index * 10 + 1] = []
            self.results.pushover_element_force[index * 10 + 2] = []

        ok = 0
        currentDisp = 0.0
        while ok == 0 and currentDisp < 0.1 * max(self.model.storey_levels):
            ok = opensees.analyze(1)
            # if the analyses fails try initial tangent __iteration
            if ok != 0:
                print("modified newton failed")
                # break
                print("regular newton failed .. lets try an KrylovNewton for this step")
                opensees.test('NormDispIncr', 1.0e-6,  1000)
                opensees.algorithm('KrylovNewton')
                ok = opensees.analyze(1)
                if ok == 0:
                    print("that worked .. back to regular newton")
                    opensees.test('NormDispIncr', 1e-6, 1000)
                    opensees.algorithm('Newton')


            currentDisp = opensees.nodeDisp(roof_node, 1)
            roof_displacements.append(currentDisp)

            for index, row in _df_elements.iterrows():
                self.results.pushover_element_deformation[index * 10 + 1].append(opensees.eleResponse(int(index * 10 + 1), 'deformation', 1))
                self.results.pushover_element_deformation[index * 10 + 2].append(opensees.eleResponse(int(index * 10 + 2), 'deformation', 1))
                self.results.pushover_element_force[index * 10 + 1].append(opensees.eleResponse(int(index * 10 + 1), 'force', 1)[5])
                self.results.pushover_element_force[index * 10 + 2].append(opensees.eleResponse(int(index * 10 + 2), 'force', 1)[5])




            _base_shear_tmp = 0.
            for i in self.model.base_level_nodes:
                _base_shear_tmp += opensees.nodeReaction(int(i)*10+1, 1)
            base_shears.append(_base_shear_tmp)


        # self.results_test['base_shears']= base_shears
        # self.results_test['roof_displacements']= roof_displacements

        # print(base_shears)
        # print(self.results_test['base_shears'])


        # print(elem_12221_mom)
        # print(elem_12221_rot)

            # print(f'δ={currentDisp} ok={ok}')
        # opensees.analyze(1500)
        # print(roof_displacements)
        # print(base_shears)

        etabs_curve = pd.read_csv(self.etabs_pushocer_curve_check, skiprows=7,
                                  sep='\s+')
        etabs_d = etabs_curve['Displacement'].tolist()
        etabs_V = etabs_curve['Base'].tolist()

        f, ax = plt.subplots(figsize=(8, 5))
        ax.plot(roof_displacements, -np.array(base_shears), label="opensees", lw=2)
        ax.plot(etabs_d, etabs_V, label="etabs", lw=2)
        # ax.axis([0, 1.3 * max(roof_displacements), 0, 1.2 * max(base_shears)])
        ax.set_title('Καμπύλη αντίστασης')
        ax.set_ylabel('V (kN)')
        ax.set_xlabel('δ (m)')
        ax.legend()
        fig = (f, ax)
        plt.show()

        # f, ax = plt.subplots(figsize=(8, 5))
        # ax.plot(elem_12221_rot, elem_12221_mom, label="M - rot", lw=2)
        # # ax.axis([0, 1.3 * max(roof_displacements), 0, 1.2 * max(base_shears)])
        # ax.set_title('Καμπύλη αντίστασης')
        # ax.set_ylabel('V (kN)')
        # ax.set_xlabel('δ (m)')
        # ax.legend()
        # fig = (f, ax)
        # plt.show()

        # print('Some pushover results')
        # print(opensees.nodeDisp(106))
        # print(opensees.eleForce(1111))


        # VsRecords = np.genfromtxt('Vbase.out', delimiter=' ', skip_footer=1).transpose()
        # Vs = VsRecords.sum(0)
        #
        # print(Vs)
        #
        # droofRecords = np.genfromtxt('droof.out', delimiter=' ', skip_footer=1)#.transpose()
        # droofs = droofRecords[0]

        # import csv
        # VsRecords = pd.read_csv('Vbase.out')
        # droofRecords = pd.read_table('droof.out', quoting=csv.QUOTE_NONE, error_bad_lines=False)
        #
        # print(droofRecords)

        # print(len(Vs))
        # print(len(droofRecords))

        # no_rows = min(len(droofs), len(Vs))

        # fig, ax = plt.subplots(figsize=(10, 7))
        # # ax.plot(droofs[:no_rows], -Vs[:no_rows])
        # ax.plot(droofRecords, -Vs)
        # ax.set(xlabel='δ roof (m)', ylabel='Base shear (kN)',
        #        title='Pushover curve')
        # ax.grid()
        # plt.show()


        opensees.wipe('all')


    def __os_run_time_history(self):

        _df_nodes = self.model.input_data['Nodes']
        _df_elements = self.model.input_data['Elements']

        roof_node = max(self.model.master_nodes)


        # w1 = 2*math.pi / 0.4
        # w2 = 2*math.pi / 0.2

        n_mIK = 10.

        w1 = 2*math.pi / self.results.elastic_periods[0]
        w2 = 2 * math.pi / self.results.elastic_periods[1]

        zeta = 0.04
        a0 = zeta*2.0*w1*w2/(w1+w2)
        a1 = zeta * 2.0 / (w1+w2)
        a1_mod = a1*(1.0+n_mIK)/n_mIK

        # elelist = _df_elements.index.tolist()
        # opensees.region(1, '-eleRange', elelist[0], elelist[-1], '-rayleigh', 0.0, 0.0, a1_mod, 0.0)
        # for mn in self.master_nodes:



        opensees.rayleigh(a0, 0.0, 0.0, a1_mod)



        opensees.recorder('Node', '-file', 'droof.out', '-node', roof_node, '-dof', 1, 'disp')
        opensees.recorder('Node', '-file', 'Vbase.out', '-node', 1011, 1021, 1031, 1041, '-dof', 1, 'reaction')

        # Ακόμα και αν δεν τυπώνω σε αρχείο, πρέπει να τρέχει το recorder για να βγάλει τις λίστες...γιατί όμως;;
        # Δεν είναι ανάγκη να υπάρχουν όλοι οι κόμβοι, φτάνει ένας, ακόμα και αν δεν είναι από αυτούς που ζητώ στη συνέχεια
        opensees.recorder('Node', '-file', 'base_level_nodes.out', '-node', self.model.base_level_nodes[0], '-dof', 1, 'reaction')




        # Set some parameters
        record = 'elCentro'

        # Permform the conversion from SMD record to OpenSees record
        dt, nPts = ReadRecord(record + '.at2', record + '.dat')


        # Set time series to be passed to uniform excitation
        opensees.timeSeries('Path', 2, '-filePath', record + '.dat', '-dt', dt, '-factor', 6.*9.81)

        # Create UniformExcitation load pattern
        #                         tag dir
        opensees.pattern('UniformExcitation', 2, 1, '-accel', 2)

        print("zzz3")

        opensees.wipeAnalysis()
        opensees.pattern('Plain', 200, 2)
        opensees.constraints('Plain')
        opensees.numberer('RCM')
        opensees.system('UmfPack')
        opensees.test('NormDispIncr', 1e-5, 5500)
        opensees.algorithm('NewtonLineSearch')
        opensees.integrator('Newmark', 0.5, 0.25)
        opensees.analysis('Transient')



        roof_displacements = [0]
        base_shears = [0]
        for index, row in _df_elements.iterrows():
            self.results.pushover_element_deformation[index * 10 + 1] = []
            self.results.pushover_element_deformation[index * 10 + 2] = []
            self.results.pushover_element_force[index * 10 + 1] = []
            self.results.pushover_element_force[index * 10 + 2] = []

        tFinal = nPts * dt
        tCurrent = opensees.getTime()

        ok = 0
        time = [tCurrent]
        u3 = [0.0]

        # Perform the transient analyses
        while ok == 0 and tCurrent < tFinal:

            ok = opensees.analyze(1, .001)

            # if the analyses fails try initial tangent __iteration
            if ok != 0:
                print("regular newton failed .. lets try an initail stiffness for this step")
                opensees.test('NormDispIncr', 1.0e-5, 5500)
                opensees.algorithm('ModifiedNewton', '-initial')
                ok = opensees.analyze(1, .001)
                if ok == 0:
                    print("that worked .. back to regular newton")
                opensees.test('NormDispIncr', 1.0e-5, 5500)
                opensees.algorithm('NewtonLineSearch')

            tCurrent =opensees.getTime()

            time.append(tCurrent)
            u3.append(opensees.nodeDisp(109, 1))

            currentDisp = opensees.nodeDisp(roof_node, 1)
            roof_displacements.append(currentDisp)

            for index, row in _df_elements.iterrows():
                self.results.pushover_element_deformation[index * 10 + 1].append(opensees.eleResponse(int(index * 10 + 1), 'deformation', 1))
                self.results.pushover_element_deformation[index * 10 + 2].append(opensees.eleResponse(int(index * 10 + 2), 'deformation', 1))
                self.results.pushover_element_force[index * 10 + 1].append(opensees.eleResponse(int(index * 10 + 1), 'force', 1)[5])
                self.results.pushover_element_force[index * 10 + 2].append(opensees.eleResponse(int(index * 10 + 2), 'force', 1)[5])

            print(tCurrent)



        print("zz4z")
        # ok = 0
        # currentDisp = 0.0
        # while ok == 0:
        #     ok = opensees.analyze(1, dt)
        #     # if the analyses fails try initial tangent __iteration
        #     if ok != 0:
        #         print("modified newton failed")
        #         # break
        #         print("regular newton failed .. lets try an KrylovNewton for this step")
        #         opensees.test('NormDispIncr', 1.0e-5,  1500)
        #         opensees.algorithm('KrylovNewton')
        #         ok = opensees.analyze(1, dt)
        #         if ok == 0:
        #             print("that worked .. back to regular NewtonLineSearch")
        #             opensees.test('NormDispIncr', 1e-5, 1500)
        #             opensees.algorithm('NewtonLineSearch')
        #
        #
        #     currentDisp = opensees.nodeDisp(roof_node, 1)
        #     roof_displacements.append(currentDisp)
        #
        #     for index, row in _df_elements.iterrows():
        #         self.results.pushover_element_deformation[index * 10 + 1].append(opensees.eleResponse(int(index * 10 + 1), 'deformation', 1))
        #         self.results.pushover_element_deformation[index * 10 + 2].append(opensees.eleResponse(int(index * 10 + 2), 'deformation', 1))
        #         self.results.pushover_element_force[index * 10 + 1].append(opensees.eleResponse(int(index * 10 + 1), 'force', 1)[5])
        #         self.results.pushover_element_force[index * 10 + 2].append(opensees.eleResponse(int(index * 10 + 2), 'force', 1)[5])
        #
        #
        #
        #
        #     _base_shear_tmp = 0.
        #     for i in self.base_level_nodes:
        #         _base_shear_tmp += opensees.nodeReaction(int(i)*10+1, 1)
        #     base_shears.append(_base_shear_tmp)






    def save_results(self):
        file = open('results.pickled', 'wb')
        # dump information to that file
        pickle.dump(self, file)
        # close the file
        file.close()

        f = gzip.open('results.gzpickled', 'wb', compresslevel=9)
        pickle.dump(self, f)
        f.close()















"""
EC8 detailing functions for beams

    .. uml::

        class beams <<(M,#FF7700)>> {
        .. functions ..
        + ρmin(fck, fyk)
        + ρmax(ρ2, μφ, εsyd, fcd, fyd)
        + ρmax05(μφ, εsyd, fcd, fyd)
        }
"""

from .....ec2.raw.ch3.concrete import strength as conc_strength


def ρmin(fck, fyk):
    return 0.5 * conc_strength.fctm(fck) / fyk


def ρmax(ρ2, μφ, εsyd, fcd, fyd):
    return ρ2 + 0.0018 * fcd / (μφ * εsyd * fyd)


def ρmax05(μφ, εsyd, fcd, fyd):
    return 2 * 0.0018 * fcd / (μφ * εsyd * fyd)



import math


def VRcalc(h, x, Ls, N, Ac, fc, μθpl, ρtot, αs, Vw):
    """ Διατµητική αντοχή δοµικού στοιχείου, όπως αυτή καθορίζεται από τη διαρροή των συνδετήρων.
    Μειώνεται συναρτήσει του μθ,pl
    Η σχέση θέλει MN και m. Αποτέλεσμα σε MN

    Args:
        h (float): ύψος διατοµής [m]
        x (float): ύψος της θλιβόµενης ζώνης [m]
        Ls (float): [m]
        N (float): Αξονικό φορτίο (θετικό για θλίψη, µηδενικό για εφελκυσµό) [ΜΝ]
        Ac (float):
        fc (float):
        μθpl (float):
        ρtot (float):
        αs (float):
        Vw (float):

    Returns:
        float: a

    """
    if N < 0:
        N = 0
    part1 = ((h - x) / (2 * Ls)) * min(N, 0.55 * Ac * fc)
    part2 = 1 - 0.05 * min(5.0, μθpl)
    part3 = 0.16 * max(0.5, 100*ρtot) * (1 - 0.16 * min(5.0, αs)) * math.pow(fc, 0.5) * Ac + Vw
    return part1 + part2 * part3


# Τέμνουσα που παραλαμβάνουν οι συνδετήρες
def Vwcalc(ρw, bw, z, fyw):

    """ Συµβολή εγκάρσιου οπλισµού στη διατµητική αντοχή

    Args:
        ρw (float): το ποσοστό του εγκάρσιου οπλισµού.
        bw (float): το πλάτος της διατομής [m]
        z (float): ο µοχλοβραχίονας εσωτερικών δυνάµεων (ίσος µε d-d’ σε υποστυλώµατα, δοκούς
                    και τοιχώµατα διατοµής Τ ή Η, ή µε 0.8h σε τοιχώµατα ορθογωνικής διατοµής) [m]
        fyw (float): η τάση διαρροής του εγκάρσιου οπλισµού [MPa]

    Returns:
        float: a

    """
    return ρw * bw * z * fyw


def VRccalc(ρtot, b, d, fc, N, Ac):
    # fc [MPa]
    # N [kN] θετικ΄η η θλ΄ίψη
    # διαστάσεις σε [m]
    v1 = max(180. * (100. * ρtot) ** (1. / 3.),
             35. * (1. + (0.2 / d) ** 0.5) ** 0.5 * fc ** (1. / 6.)
             )

    return (v1 * (1 + (0.2 / d) ** 0.5) * fc ** (1 / 3) + 0.15 * N / Ac) * b * d


"""

    .. uml::

        class combos <<(M,#FF7700)>> {
        .. classes ..
        LongReinforcementLayer
        TransReinforcementLayer
        }

"""

from dataclasses import dataclass, field
from typing import List
import numpy as np

from .areas import As_layer


@dataclass
class LongReinforcementLayer:
    ns: List[float]  # = field(default_factory=list)
    dias: List[float]  # = field(default_factory=list)

    @classmethod
    def from_string(cls, reinf_string, units_input='mm', units_output='mm', dia_symbol='Φ'):
        # Πχ reinf_string='2Φ18+2Φ16'

        # Μετατρέπω σε mm (συνηθέστερο) για να γίνουν οι πράξεις
        if units_input == 'm':
            length_multiplier_input = 1000.
        elif units_input == 'cm':
            length_multiplier_input = 10.
        else:
            length_multiplier_input = 1.

        ns_and_dias = [x.split(dia_symbol) for x in reinf_string.split('+')]
        ns = [float(y[0]) for y in ns_and_dias]
        dias = [float(y[1]) * length_multiplier_input for y in ns_and_dias]

        if units_output == 'm':
            length_multiplier_output = 0.001
        elif units_output == 'cm':
            length_multiplier_output = 0.1
        else:
            length_multiplier_output = 1.0

        return cls(ns, [d*length_multiplier_output for d in dias])

    def __post_init__(self):
        self.__Astot = As_layer(self.ns, self.dias)
        self.__ntot = sum(self.ns)
        self.__dia_max = max(self.dias)
        self.__dia_min = min(self.dias)
        if self.ntot > 0:
            self.__dia_equiv = np.sqrt(4*self.Astot / (self.ntot * np.pi))
        else:
            self.__dia_equiv = 0.0
        self.__As_equiv = np.pi * self.__dia_equiv**2 / 4

    @property
    def Astot(self):
        return self.__Astot

    @property
    def ntot(self):
        return self.__ntot

    @property
    def dia_max(self):
        return self.__dia_max

    @property
    def dia_min(self):
        return self.__dia_min

    @property
    def dia_equiv(self):
        return self.__dia_equiv

    @property
    def As_equiv(self):
        return self.__As_equiv


@dataclass
class TransReinforcementLayer:
    n: int
    dia: float
    s: float

    @classmethod
    def from_string(cls, reinf_string, units_input='mm', units_output='mm', dia_symbol='Φ'):
        # Πχ reinf_string='Φ8/140(3)'

        # Μετατρέπω σε mm (συνηθέστερο) για να γίνουν οι πράξεις
        if units_input == 'm':
            length_multiplier_input = 1000.
        elif units_input == 'cm':
            length_multiplier_input = 10.
        else:
            length_multiplier_input = 1.

        n = int(reinf_string[reinf_string.find("(")+1:reinf_string.find(")")])
        dia = float(reinf_string[reinf_string.find(
            dia_symbol)+1:reinf_string.find("/")]) * length_multiplier_input
        s = float(reinf_string[reinf_string.find(
            "/")+1:reinf_string.find("(")]) * length_multiplier_input

        if units_output == 'm':
            length_multiplier_output = 0.001
        elif units_output == 'cm':
            length_multiplier_output = 0.1
        else:
            length_multiplier_output = 1.0

        return cls(n, dia * length_multiplier_output, s * length_multiplier_output)

    @property
    def As(self):
        return As_layer(self.n, self.dia)



"""

    .. uml::

        class areas <<(M,#FF7700)>> {
        .. functions ..
        As()
        As_distributed()
        As_for_ω()
        As_layer()
        As1_for_ω1ρ1()
        As2_for_ω2ρ2()
        }

"""

import numpy as np


def As(dia):
    return np.pi * dia**2 / 4


def As_distributed(dia, s):
    """
    Area reinforcement for 1.0m length

    Args:
        dia (float): Bar diameter
        s (float): Distance between bars

    Returns:
        float:

    """

    return As(dia) * 1.0 / s


def As_layer(n, dia):
    """

    Args:
        n (list or integer):
        dia (np.array or float):

    Returns:
        float:
    """

    if isinstance(n, list):
        return sum(np.array(n) * np.array(dia)**2 * np.pi/4)
    else:
        return n * dia**2 * np.pi/4


def As_for_ω(ω, b, d, fc, fy, N):
    """

    Args:
        ω (float):
        b (float): Section width
        d (float): Section effective depth
        fc (float): Concrete compressive strength
        fy (float): Reinforcement steel yield strength
        N (float): Axial force

    Returns:
        float:

    """
    return ω * b * d * fc / fy + N / fy


def As1_for_ω1ρ1(ω1, ρ1, b, d, fc, fy, N):
    return ω1 * ρ1 * b * d * fc / fy + N / fy


def As2_for_ω2ρ2(ω2, ρ2, b, d, fc, fy):
    return ω2 * ρ2 * b * d * fc / fy



import pandas as pd
import numpy as np
from dataclasses import dataclass, field
from typing import Tuple, Dict, List
from ....common.math import numerical
from ....common.math.charts.linexy import LineXY


@dataclass
class DeriveProcedure:
    excel_filename: str
    excel_worksheet: str
    raw_data: pd.DataFrame = None
    base_records: np.ndarray = None
    pgas_unique: np.ndarray = None
    all_raw_curves: Dict = field(default_factory=dict)
    all_L_values: List = field(default_factory=dict)
    average_L_values: np.ndarray = None
    thresholds: np.ndarray = None

    pga_at_threshold_for_base_record: List = field(default_factory=dict)
    pga_at_threshold_for_damage_state: List = field(default_factory=dict)

    pga_for_thresholds: np.ndarray = None

    def __post_init__(self):
        self.raw_data = pd.read_excel(self.excel_filename, self.excel_worksheet)
        self.base_records = self.raw_data['base_record'].unique()
        self.pgas_unique = np.sort(self.raw_data['pga'].unique())



    def calc(self):
        self.all_raw_curves = {}
        self.all_L_values = []
        for rec_name in self.base_records:
            tmp_df = self.raw_data_for_single_base_record(rec_name)
            xyxhart = LineXY(x=np.array(tmp_df['pga'].tolist()),
                             y=np.array(tmp_df['loss_index'].tolist()),
                             name=rec_name)
            self.all_raw_curves[rec_name] = xyxhart
            self.all_L_values.append(xyxhart.y)

        self.average_L_values = np.average(self.all_L_values, axis=0)

        self.pga_for_thresholds = self.get_pga_for_thresholds(self.thresholds, self.average_L_values, self.pgas_unique)

        self.pga_at_threshold_for_base_record = []
        for i in self.all_L_values:
            self.pga_at_threshold_for_base_record.append(self.get_pga_for_thresholds(self.thresholds,
                                                                                     i,
                                                                                     self.pgas_unique))

        self.pga_at_threshold_for_damage_state = np.transpose(self.pga_at_threshold_for_base_record)


    @staticmethod
    def get_pga_for_thresholds(_thresholds, _L_values, _pgas):
        return np.interp(_thresholds, _L_values, _pgas)

    def raw_data_for_single_base_record(self, base_record):
        return self.raw_data[self.raw_data.base_record==base_record]
